# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np

from thermohl import solver
from thermohl.solver import HeatEquationType
from thermohl.solver.entities import (
    CableType,
    ModelType,
    VariableType,
    TemperatureType,
    TargetType,
)

_nprs = 123456


def _solvers(dic=None):
    return [
        solver._factory(dic=dic, heat_equation=heat_equation, model=m)
        for heat_equation in [
            HeatEquationType.THREE_TEMPERATURES,
            HeatEquationType.THREE_TEMPERATURES_LEGACY,
        ]
        for m in [
            ModelType.RTE,
            ModelType.CIGRE,
            ModelType.IEEE,
            ModelType.OLLA,
        ]
    ]


def test_balance():
    # NB : this one fails only with 'cigre' powers

    tol = 1.0e-09

    rng = np.random.default_rng(_nprs)

    N = 9999
    dic = {
        "latitude": rng.random(N) * 9 + 42.0,
        "altitude": rng.random(N) * 1600.0,
        "cable_azimuth": rng.random(N) * 360.0,
        "month": rng.integers(1, 13, N),
        "day": rng.integers(1, 31, N),
        "hour": rng.integers(0, 24, N),
        "ambient_temperature": rng.random(N) * 30.0,
        "wind_speed": rng.random(N) * 7.0,
        "wind_azimuth": rng.random(N) * 90.0,
        "transit": rng.random(N) * 3960 + 40.0,
        "core_diameter": rng.integers(2, size=N)
        * solver.default_values()["core_diameter"],
    }

    for s in _solvers(dic):
        # compute guess with 1t solver
        s1 = solver._factory(
            dic=dic,
            heat_equation=HeatEquationType.ONE_TEMPERATURE,
            model=ModelType.IEEE,
        )
        t1 = s1.steady_temperature(
            tol=2.0, maxiter=16, return_err=False, return_power=False
        )
        t1 = t1[VariableType.TEMPERATURE].values
        # 3t solve
        df = s.steady_temperature(
            surface_temperature_guess=t1,
            core_temperature_guess=t1,
            return_err=True,
            return_power=True,
            tol=tol,
            maxiter=64,
        )
        # checks
        assert np.all(df[VariableType.ERROR] < tol)
        assert np.allclose(
            s.balance(
                surface_temperature=df[TemperatureType.SURFACE],
                core_temperature=df[TemperatureType.CORE],
            ).values,
            0.0,
            atol=tol,
        )
        assert np.allclose(
            s.morgan(
                surface_temperature=df[TemperatureType.SURFACE],
                core_temperature=df[TemperatureType.CORE],
            ).values,
            0.0,
            atol=tol,
        )


def test_consistency():
    tol = 1.0e-09

    rng = np.random.default_rng(_nprs)

    N = 9999

    dic = {
        "latitude": rng.random(N) * 9 + 42.0,
        "altitude": rng.random(N) * 1600.0,
        "cable_azimuth": rng.random(N) * 360.0,
        "month": rng.integers(1, 13, N),
        "day": rng.integers(1, 31, N),
        "hour": rng.integers(0, 24, N),
        "ambient_temperature": rng.random(N) * 30.0,
        "wind_speed": rng.random(N) * 7.0,
        "wind_azimuth": rng.random(N) * 90.0,
        "core_diameter": rng.integers(2, size=N)
        * solver.default_values()["core_diameter"],
    }

    for s in _solvers(dic):
        d = {
            TargetType.SURFACE: TemperatureType.SURFACE,
            TargetType.AVERAGE: TemperatureType.AVERAGE,
            TargetType.CORE: TemperatureType.CORE,
        }

        for location, temperature_at_location in d.items():
            # solve intensity with different targets
            df = s.steady_intensity(
                max_conductor_temperature=100.0,
                target=location,
                return_err=True,
                return_power=True,
                tol=1.0e-09,
                maxiter=64,
            )
            assert np.all(df[VariableType.ERROR] < tol)
            # set args intensity to newly founds ampacities
            s.args.I = df[VariableType.TRANSIT].values
            s.update()
            assert np.allclose(
                s.balance(
                    surface_temperature=df[TemperatureType.SURFACE],
                    core_temperature=df[TemperatureType.CORE],
                ).values,
                0.0,
                atol=tol,
            )
            assert np.allclose(
                s.morgan(
                    surface_temperature=df[TemperatureType.SURFACE],
                    core_temperature=df[TemperatureType.CORE],
                ).values,
                0.0,
                atol=tol,
            )
            # 3t solve
            dg = s.steady_temperature(
                surface_temperature_guess=df[TemperatureType.SURFACE].round(1).values,
                core_temperature_guess=df[TemperatureType.CORE].round(1).values,
                return_err=True,
                return_power=True,
                tol=1.0e-09,
                maxiter=64,
            )
            # check consistency
            assert np.allclose(dg[temperature_at_location].values, 100.0)


def test_steady_intensity_cable_type_scalar_homogeneous():
    for s in _solvers():
        df_with_specified_cable_type = s.steady_intensity(
            max_conductor_temperature=100.0,
            cable_type=CableType.HOMOGENEOUS,
            return_err=True,
            return_power=True,
        )

        df_with_specified_target = s.steady_intensity(
            max_conductor_temperature=100.0,
            target=TargetType.AVERAGE,
            return_err=True,
            return_power=True,
        )

        assert np.allclose(df_with_specified_cable_type, df_with_specified_target)


def test_steady_intensity_cable_type_scalar_bimetallic():
    for s in _solvers():
        df_with_specified_cable_type = s.steady_intensity(
            max_conductor_temperature=100.0,
            cable_type=CableType.BIMETALLIC,
            return_err=True,
            return_power=True,
        )

        df_with_specified_target = s.steady_intensity(
            max_conductor_temperature=100.0,
            target=TargetType.CORE,
            return_err=True,
            return_power=True,
        )

        assert np.allclose(df_with_specified_cable_type, df_with_specified_target)


def test_steady_intensity_cable_type_list():
    rng = np.random.default_rng(_nprs)

    N = 3

    dic = {
        "latitude": rng.random(N) * 9 + 42.0,
        "altitude": rng.random(N) * 1600.0,
        "azimuth": rng.random(N) * 360.0,
        "month": rng.integers(1, 13, N),
        "day": rng.integers(1, 31, N),
        "hour": rng.integers(0, 24, N),
        "ambient_temperature": rng.random(N) * 30.0,
        "wind_speed": rng.random(N) * 7.0,
        "wind_angle": rng.random(N) * 90.0,
        "core_diameter": rng.integers(2, size=N)
        * solver.default_values()["core_diameter"],
    }

    cable_type = np.array(
        [CableType.BIMETALLIC, CableType.HOMOGENEOUS, CableType.HOMOGENEOUS]
    )
    target = np.array([TargetType.CORE, TargetType.AVERAGE, TargetType.AVERAGE])

    for s in _solvers(dic):
        df_with_specified_cable_type = s.steady_intensity(
            max_conductor_temperature=100.0,
            cable_type=cable_type,
            return_err=True,
            return_power=True,
        )

        df_with_specified_target = s.steady_intensity(
            max_conductor_temperature=100.0,
            target=target,
            return_err=True,
            return_power=True,
        )

        assert np.allclose(df_with_specified_cable_type, df_with_specified_target)


def test_steady_intensity_cable_type_and_target():
    # Both cable_type and target are provided so target should be ignored.
    rng = np.random.default_rng(_nprs)

    N = 9999
    dic = {
        "latitude": rng.random(N) * 9 + 42.0,
        "altitude": rng.random(N) * 1600.0,
        "azimuth": rng.random(N) * 360.0,
        "month": rng.integers(1, 13, N),
        "day": rng.integers(1, 31, N),
        "hour": rng.integers(0, 24, N),
        "ambient_temperature": rng.random(N) * 30.0,
        "wind_speed": rng.random(N) * 7.0,
        "wind_angle": rng.random(N) * 90.0,
        "core_diameter": rng.integers(2, size=N)
        * solver.default_values()["core_diameter"],
    }

    cable_type = rng.choice([CableType.BIMETALLIC, CableType.HOMOGENEOUS], size=N)

    target = np.array([TargetType.CORE] * N)

    for s in _solvers(dic):
        df_with_specified_cable_type = s.steady_intensity(
            max_conductor_temperature=100.0,
            cable_type=cable_type,
            target=target,
            return_err=True,
            return_power=True,
        )

        df_with_specified_target = s.steady_intensity(
            max_conductor_temperature=100.0,
            cable_type=cable_type,
            return_err=True,
            return_power=True,
        )

        assert np.allclose(df_with_specified_cable_type, df_with_specified_target)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timezone
import numpy as np

from thermohl.power import ieee
from thermohl import solver


def test_compare_powers():
    """Compare computed values to hard-coded ones from ieee std 38-2012."""
    dic = solver.default_values()

    # there are a lot of rounding in the standard guide, hence the relatively
    # large tolerances used in our tests ...

    dic["wind_speed"] = 0.61
    dic["wind_azimuth"] = 0.0
    dic["emissivity"] = 0.8
    dic["solar_absorptivity"] = 0.8
    dic["ambient_temperature"] = 40.0
    dic["temp_high"] = 75.0
    dic["temp_low"] = 25.0
    dic["linear_resistance_temp_high"] = 8.688e-05
    dic["linear_resistance_temp_low"] = 7.283e-05
    dic["cable_azimuth"] = 90.0
    dic["latitude"] = 30.0
    dic["turbidity"] = 0.0
    dic["altitude"] = 0.0
    dic["outer_diameter"] = 28.14 * 1.0e-03
    dic["core_diameter"] = 10.4 * 1.0e-03
    dic["datetime_utc"] = datetime(2000, 6, 10, 11, tzinfo=timezone.utc)

    conductor_temperature = 100.0

    assert np.isclose(
        ieee.ConvectiveCooling(**dic).value(conductor_temperature), 81.93, rtol=0.002
    )
    assert np.isclose(
        ieee.RadiativeCooling(**dic).value(conductor_temperature), 39.1, rtol=0.001
    )
    assert np.isclose(
        ieee.SolarHeating(**dic).value(conductor_temperature), 22.44, rtol=0.006
    )
    joule_heating = ieee.JouleHeating(**dic)
    assert np.isclose(
        joule_heating._rdc(conductor_temperature), 9.390e-05, rtol=1.0e-09
    )

    # additional debug
    ieee.SolarHeating(**dic).value(conductor_temperature)

    from thermohl import sun

    sd = sun.solar_declination(dic["datetime_utc"].date())
    assert np.isclose(np.rad2deg(sd), 23.0, atol=0.1)

    ha = sun.hour_angle(sun.time_to_float_hours(dic["datetime_utc"].time()))
    assert np.isclose(np.rad2deg(ha), -15.0, atol=0.1)

    sa = sun.solar_altitude(
        np.deg2rad(dic["latitude"]),
        dic["datetime_utc"].date(),
        sun.time_to_float_hours(dic["datetime_utc"].time()),
    )
    assert np.isclose(np.rad2deg(sa), 74.8, atol=0.2)

    sz = sun.solar_azimuth(
        np.deg2rad(dic["latitude"]),
        dic["datetime_utc"].date(),
        sun.time_to_float_hours(dic["datetime_utc"].time()),
    )
    assert np.isclose(np.rad2deg(sz), 114.0, atol=0.5)

    th = np.arccos(np.cos(sa) * np.cos(sz - np.deg2rad(dic["cable_azimuth"])))
    assert np.isclose(np.rad2deg(th), 76.1, atol=1.5)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timezone
import pytest
import numpy as np
import pandas as pd

from thermohl import power
from thermohl.solver.entities import PowerType, VariableType
from thermohl.solver.slv1t import Solver1T


@pytest.fixture
def solver():
    args = {
        "max_len": lambda: 1,
        "transit": np.array([0]),
        "ambient_temperature": np.array([25]),
        "wind_speed": np.array([0]),
        "wind_azimuth": np.array([0]),
        "ambient_pressure": np.array([101325]),
        "relative_humidity": np.array([50]),
        "precipitation_rate": np.array([0]),
        "linear_mass": 1.0,
        "heat_capacity": 1.0,
        "datetime_utc": datetime(2025, 1, 1, 0, tzinfo=timezone.utc),
        "latitude": np.array([48.0]),
        "longitude": np.array([2.3]),
        "altitude": np.array([50.0]),
        "cable_azimuth": np.array([0.0]),
        "solar_absorptivity": np.array([0.5]),
    }
    cable = {
        "outer_diameter": np.array([3.186e-02]),
        "core_diameter": np.array([0.000]),
        "outer_area": np.array([6.004e-4]),
        "core_area": np.array([0.000]),
        "magnetic_coeff": np.array([1.000]),
        "magnetic_coeff_per_a": np.array([0.000]),
        "temperature_coeff_linear": np.array([3.600e-3]),
        "temperature_coeff_quadratic": np.array([8.000e-7]),
        "linear_resistance_dc_20c": np.array([5.540e-5]),
        "emissivity": np.array([0.8]),
    }
    args.update(cable)

    joule = power.rte.joule_heating.JouleHeating
    solar = power.rte.solar_heating.SolarHeating
    convective = power.rte.convective_cooling.ConvectiveCooling
    radiative = power.rte.radiative_cooling.RadiativeCooling

    solver = Solver1T(
        dic=args, joule=joule, solar=solar, convective=convective, radiative=radiative
    )
    return solver


def test_steady_temperature_default(solver):
    result = solver.steady_temperature()

    assert isinstance(result, pd.DataFrame)
    assert VariableType.TEMPERATURE in result.columns
    assert PowerType.JOULE in result.columns
    assert PowerType.SOLAR in result.columns
    assert PowerType.CONVECTION in result.columns
    assert PowerType.RADIATION in result.columns
    assert PowerType.RAIN in result.columns


def test_steady_temperature_with_error(solver):
    result = solver.steady_temperature(return_err=True)

    assert isinstance(result, pd.DataFrame)
    assert VariableType.TEMPERATURE in result.columns
    assert VariableType.ERROR in result.columns


def test_steady_temperature_no_power(solver):
    result = solver.steady_temperature(return_power=False)

    assert isinstance(result, pd.DataFrame)
    assert VariableType.TEMPERATURE in result.columns
    assert PowerType.JOULE not in result.columns
    assert PowerType.SOLAR not in result.columns
    assert PowerType.CONVECTION not in result.columns
    assert PowerType.RADIATION not in result.columns
    assert PowerType.RAIN not in result.columns


def test_steady_temperature_custom_params(solver):
    Tmin = 10.0
    Tmax = 50.0
    tol = 1e-5
    maxiter = 100

    result = solver.steady_temperature(
        Tmin=Tmin,
        Tmax=Tmax,
        tol=tol,
        maxiter=maxiter,
    )

    assert isinstance(result, pd.DataFrame)
    assert VariableType.TEMPERATURE in result.columns


def test_transient_temperature_default(solver):
    time = np.array([0, 1, 2, 3, 4, 5])

    result = solver.transient_temperature(time)

    assert isinstance(result, dict)
    assert VariableType.TIME in result
    assert VariableType.TEMPERATURE in result
    assert len(result[VariableType.TIME]) == len(time)
    assert len(result[VariableType.TEMPERATURE]) == len(time)


def test_transient_temperature_with_initial_temp(solver):
    time = np.array([0, 1, 2, 3, 4, 5])
    T0 = 30.0

    result = solver.transient_temperature(time, T0=T0)

    assert isinstance(result, dict)
    assert VariableType.TIME in result
    assert VariableType.TEMPERATURE in result
    assert len(result[VariableType.TIME]) == len(time)
    assert len(result[VariableType.TEMPERATURE]) == len(time)
    assert result[VariableType.TEMPERATURE][0] == T0


def test_transient_temperature_with_error(solver):
    time = np.array([0, 1, 2, 3, 4, 5])

    result = solver.transient_temperature(time, return_power=True)

    assert isinstance(result, dict)
    assert VariableType.TIME in result
    assert VariableType.TEMPERATURE in result
    assert len(result[VariableType.TIME]) == len(time)
    assert len(result[VariableType.TEMPERATURE]) == len(time)
    assert PowerType.JOULE in result
    assert PowerType.SOLAR in result
    assert PowerType.CONVECTION in result
    assert PowerType.RADIATION in result
    assert PowerType.RAIN in result


def test_steady_intensity_default(solver):
    conductor_temperature = np.array([75])

    result = solver.steady_intensity(conductor_temperature)

    assert isinstance(result, pd.DataFrame)
    assert VariableType.TRANSIT in result.columns
    assert PowerType.JOULE in result.columns
    assert PowerType.SOLAR in result.columns
    assert PowerType.CONVECTION in result.columns
    assert PowerType.RADIATION in result.columns
    assert PowerType.RAIN in result.columns


def test_steady_intensity_with_error(solver):
    conductor_temperature = np.array([75])

    result = solver.steady_intensity(conductor_temperature, return_err=True)

    assert isinstance(result, pd.DataFrame)
    assert VariableType.TRANSIT in result.columns
    assert VariableType.ERROR in result.columns


def test_steady_intensity_no_power(solver):
    conductor_temperature = np.array([75])

    result = solver.steady_intensity(conductor_temperature, return_power=False)

    assert isinstance(result, pd.DataFrame)
    assert VariableType.TRANSIT in result.columns
    assert PowerType.JOULE not in result.columns
    assert PowerType.SOLAR not in result.columns
    assert PowerType.CONVECTION not in result.columns
    assert PowerType.RADIATION not in result.columns
    assert PowerType.RAIN not in result.columns


def test_steady_intensity_custom_params(solver):
    conductor_temperature = np.array([75])
    Imin = 5.0
    Imax = 1010.0
    tol = 1e-5
    maxiter = 100

    result = solver.steady_intensity(
        conductor_temperature, Imin=Imin, Imax=Imax, tol=tol, maxiter=maxiter
    )

    assert isinstance(result, pd.DataFrame)
    assert VariableType.TRANSIT in result.columns



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any


from thermohl import floatArrayLike, datetimeListLike
from thermohl.power import _SRad, SolarHeatingBase


class SolarHeating(SolarHeatingBase):
    def __init__(
        self,
        latitude: floatArrayLike,
        altitude: floatArrayLike,
        cable_azimuth: floatArrayLike,
        turbidity: floatArrayLike,
        datetime_utc: datetimeListLike,
        outer_diameter: floatArrayLike,
        solar_absorptivity: floatArrayLike,
        measured_solar_irradiance: floatArrayLike,
        **kwargs: Any,
    ):
        """Init with args.
        If more than one input are numpy arrays, they should have the same size.

        :param latitude: Latitude in degrees.
        :param altitude: Altitude.
        :param cable_azimuth: Azimuth of the conductor in degrees.
        :param turbidity: Air pollution from 0 (clean) to 1 (polluted).
        :param datetime_utc: Datetime in UTC.
        :param outer_diameter: external diameter of the conductor.
        :param solar_absorptivity: Solar absorption coefficient of the conductor.
        :param measured_solar_irradiance: Optional precomputed solar radiation term.
        """
        est = _SRad(
            [
                -4.22391e01,
                +6.38044e01,
                -1.9220e00,
                +3.46921e-02,
                -3.61118e-04,
                +1.94318e-06,
                -4.07608e-09,
            ],
            [
                +5.31821e01,
                +1.4211e01,
                +6.6138e-01,
                -3.1658e-02,
                +5.4654e-04,
                -4.3446e-06,
                +1.3236e-08,
            ],
        )
        super().__init__(
            latitude,
            altitude,
            cable_azimuth,
            turbidity,
            datetime_utc,
            outer_diameter,
            solar_absorptivity,
            est,
            measured_solar_irradiance,
            **kwargs,
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np
from thermohl.power.rte.solar_heating import compute_solar_irradiance


def test_compute_solar_irradiance_night():
    # When solar_altitude <= 0, irradiance should be 0
    res = compute_solar_irradiance(
        global_radiation=500.0,
        solar_altitude=-0.1,
        incidence=0.5,
        nebulosity=4.0,
        albedo=0.2,
    )
    assert res == 0.0

    res = compute_solar_irradiance(
        global_radiation=500.0,
        solar_altitude=0.0,
        incidence=0.5,
        nebulosity=4.0,
        albedo=0.2,
    )
    assert res == 0.0


def test_compute_solar_irradiance_day():
    global_radiation = 800.0
    solar_altitude = 0.5  # radians
    incidence = 0.8  # radians
    nebulosity = 2.0
    albedo = 0.15

    # Manual calculation
    # diffuse_radiation = 800 * (0.3 + 0.7 * (2/8)**2) = 275.0
    # beam_radiation = (800 - 275) / sin(0.5) = 251.6984077672066
    # solar_irradiance = beam_radiation * (sin(0.8) + pi/2 * 0.15 * sin(0.5)) + 275 * pi/2 * (1 + 0.15) = 1406.012913525969
    expected = 1406.012913525969

    res = compute_solar_irradiance(
        global_radiation=global_radiation,
        solar_altitude=solar_altitude,
        incidence=incidence,
        nebulosity=nebulosity,
        albedo=albedo,
    )

    assert np.isclose(res, expected)


def test_compute_solar_irradiance_array():
    global_radiation = np.array([700.0, 400.0, 600.0])
    solar_altitude = np.array([0.4, -0.1, 0.0])
    incidence = np.array([0.6, 0.2, 0.3])
    nebulosity = np.array([3.0, 5.0, 4.0])
    albedo = np.array([0.1, 0.2, 0.15])

    res = compute_solar_irradiance(
        global_radiation=global_radiation,
        solar_altitude=solar_altitude,
        incidence=incidence,
        nebulosity=nebulosity,
        albedo=albedo,
    )

    assert len(res) == 3
    assert res[1] == 0.0
    assert res[2] == 0.0

    # Check first element
    # diffuse = 700 * (0.3 + 0.7 * (3 / 8) ** 2) = 278.90625
    # beam = (700 - 278.90625) / sin(0.4) = 1081.3340623351336
    # expected0 = beam * (sin(0.6) + pi / 2 * 0.1 * sin(0.4)) + diffuse * pi / 2 * (1.1) = 1158.648723646428
    expected0 = 1158.631321677975
    assert np.isclose(res[0], expected0)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import pytest
import numpy as np

from thermohl import solver
from thermohl.solver import ModelType
from thermohl.solver.entities import (
    HeatEquationType,
    VariableType,
    PowerType,
)
from thermohl.solver.parameters import DEFAULT_PARAMETERS as DP

_nprs = 123456


def _solvers(dic=None):
    return [
        solver._factory(
            dic=dic, heat_equation=HeatEquationType.ONE_TEMPERATURE, model=m
        )
        for m in [
            ModelType.RTE,
            ModelType.CIGRE,
            ModelType.IEEE,
            ModelType.OLLA,
        ]
    ]


def test_balance():
    tol = 1.0e-09
    np.random.seed(_nprs)
    N = 9999
    dic = dict(
        latitude=np.random.uniform(42.0, 51.0, N),
        altitude=np.random.uniform(0.0, 1600.0, N),
        cable_azimuth=np.random.uniform(0.0, 360.0, N),
        month=np.random.randint(1, 13, N),
        day=np.random.randint(1, 31, N),
        hour=np.random.randint(0, 24, N),
        ambient_temperature=np.random.uniform(0.0, 30.0, N),
        wind_speed=np.random.uniform(0.0, 7.0, N),
        wind_azimuth=np.random.uniform(0.0, 90.0, N),
        transit=np.random.uniform(40.0, 4000.0, N),
        core_diameter=np.random.randint(2, size=N)
        * solver.default_values()["core_diameter"],
    )

    for s in _solvers(dic):
        df = s.steady_temperature(
            return_err=True, return_power=True, tol=tol, maxiter=64
        )
        assert np.all(df[VariableType.ERROR] < tol)
        bl = np.abs(
            df[PowerType.JOULE]
            + df[PowerType.SOLAR]
            - df[PowerType.CONVECTION]
            - df[PowerType.RADIATION]
            - df[PowerType.RAIN]
        )
        atol = np.maximum(
            np.abs(
                s.balance(df[VariableType.TEMPERATURE] + 0.5 * df[VariableType.ERROR])
            ),
            np.abs(
                s.balance(df[VariableType.TEMPERATURE] - 0.5 * df[VariableType.ERROR])
            ),
        )
        assert np.allclose(bl, 0.0, atol=atol)


def test_consistency():
    np.random.seed(_nprs)
    N = 9999
    dic = dict(
        latitude=np.random.uniform(42.0, 51.0, N),
        altitude=np.random.uniform(0.0, 1600.0, N),
        cable_azimuth=np.random.uniform(0.0, 360.0, N),
        month=np.random.randint(1, 13, N),
        day=np.random.randint(1, 31, N),
        hour=np.random.randint(0, 24, N),
        ambient_temperature=np.random.uniform(0.0, 30.0, N),
        wind_speed=np.random.uniform(0.0, 7.0, N),
        wind_azimuth=np.random.uniform(0.0, 90.0, N),
        core_diameter=np.random.randint(2, size=N)
        * solver.default_values()["core_diameter"],
    )

    for s in _solvers(dic):
        df = s.steady_intensity(
            max_conductor_temperature=100.0,
            return_err=True,
            return_power=True,
            tol=1.0e-09,
            maxiter=64,
        )
        bl = (
            df[PowerType.JOULE]
            + df[PowerType.SOLAR]
            - df[PowerType.CONVECTION]
            - df[PowerType.RADIATION]
            - df[PowerType.RAIN]
        )
        assert np.allclose(bl, 0.0, atol=1.0e-06)
        s.args[VariableType.TRANSIT.value] = df[VariableType.TRANSIT].values
        s.update()
        dg = s.steady_temperature(
            return_err=True, return_power=True, tol=1.0e-09, maxiter=64
        )
        assert np.allclose(dg[VariableType.TEMPERATURE].values, 100.0)


def test_steady_intensity_hot_weather():
    ambient_temperature = np.array([30.0, 35.0, 40.0, 45.0, 50.0])

    solver_1t = solver.ieee(
        dic={
            "ambient_temperature": ambient_temperature,
        },
        heat_equation=HeatEquationType.ONE_TEMPERATURE,
    )

    # Here some ambient temperatures are above the maximum conductor temperature,
    # so there's no solution - the solver should raise a ValueError
    with pytest.raises(ValueError):
        solver_1t.steady_intensity(
            max_conductor_temperature=45,
            Imin=np.ones_like(ambient_temperature) * DP.imin,
            Imax=np.ones_like(ambient_temperature) * DP.imax,
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np
import pytest

from thermohl.power.cigre import JouleHeating

joule_heating_instances = [
    JouleHeating(
        transit=np.array([100.0, 150.0, 200.0]),
        magnetic_coeff=np.array([1.0, 1.0, 1.0]),
        temperature_coeff_linear=np.array([0.004, 0.004, 0.004]),
        linear_resistance_dc_20c=np.array([0.1, 0.1, 0.1]),
        reference_temperature=np.array([20.0, 18.0, 22.0]),
    ),
    JouleHeating(
        transit=100.0,
        magnetic_coeff=1.0,
        temperature_coeff_linear=0.004,
        linear_resistance_dc_20c=0.1,
        reference_temperature=20.0,
    ),
]


@pytest.mark.parametrize(
    "joule_heating",
    joule_heating_instances,
    ids=[
        "JouleHeating with arrays",
        "JouleHeating with scalars",
    ],
)
def test_joule_heating_value_scalar(joule_heating):
    conductor_temperature = 25.0
    expected = (
        joule_heating.magnetic_coeff
        * joule_heating.linear_resistance_dc_20c
        * (
            1.0
            + joule_heating.temp_coeff_linear
            * (conductor_temperature - joule_heating.reference_temperature)
        )
        * joule_heating.transit**2
    )

    result = joule_heating.value(conductor_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


@pytest.mark.parametrize(
    "joule_heating",
    joule_heating_instances,
    ids=[
        "JouleHeating with arrays",
        "JouleHeating with scalars",
    ],
)
def test_joule_heating_value_array(joule_heating):
    conductor_temperature = np.array([25.0, 30.0, 35.0])
    expected = (
        joule_heating.magnetic_coeff
        * joule_heating.linear_resistance_dc_20c
        * (
            1.0
            + joule_heating.temp_coeff_linear
            * (conductor_temperature - joule_heating.reference_temperature)
        )
        * joule_heating.transit**2
    )

    result = joule_heating.value(conductor_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_joule_heating_value_mismatched_array_sizes_should_raise_error():
    transit = np.array([100.0, 150.0])
    magnetic_coeff = np.array([1.0, 1.0, 1.0])
    temperature_coeff_linear = np.array([0.004, 0.004])
    linear_resistance_dc_20c = np.array([0.1, 0.1])
    reference_temperature = np.array([20.0, 20.0])
    conductor_temperature = np.array([25.0, 30.0])
    with pytest.raises(ValueError):
        joule_heating = JouleHeating(
            transit,
            magnetic_coeff,
            temperature_coeff_linear,
            linear_resistance_dc_20c,
            reference_temperature,
        )
        joule_heating.value(conductor_temperature)


@pytest.mark.parametrize(
    "joule_heating",
    joule_heating_instances,
    ids=[
        "JouleHeating with arrays",
        "JouleHeating with scalars",
    ],
)
def test_joule_heating_derivative_scalar(joule_heating):
    conductor_temperature = 25.0
    expected = (
        joule_heating.magnetic_coeff
        * joule_heating.linear_resistance_dc_20c
        * joule_heating.temp_coeff_linear
        * joule_heating.transit**2
    )

    result = joule_heating.derivative(conductor_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


@pytest.mark.parametrize(
    "joule_heating",
    joule_heating_instances,
    ids=[
        "JouleHeating with arrays",
        "JouleHeating with scalars",
    ],
)
def test_joule_heating_derivative_array(joule_heating):
    conductor_temperature = np.array([25.0, 30.0, 35.0])
    expected = (
        joule_heating.magnetic_coeff
        * joule_heating.linear_resistance_dc_20c
        * joule_heating.temp_coeff_linear
        * joule_heating.transit**2
    )

    result = joule_heating.derivative(conductor_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_joule_heating_derivative_mismatched_array_sizes_should_raise_error():
    transit = np.array([100.0, 150.0])
    magnetic_coeff = np.array([1.0, 1.0, 1.0])
    temperature_coeff_linear = np.array([0.004, 0.004])
    linear_resistance_dc_20c = np.array([0.1, 0.1])
    reference_temperature = np.array([20.0, 20.0])
    conductor_temperature = np.array([25.0, 30.0])
    with pytest.raises(ValueError):
        joule_heating = JouleHeating(
            transit,
            magnetic_coeff,
            temperature_coeff_linear,
            linear_resistance_dc_20c,
            reference_temperature,
        )
        joule_heating.derivative(conductor_temperature)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timezone
import numpy as np
from numpy import ndarray

from thermohl.solver.solver import reshape, _set_dates
from thermohl.solver import Parameters


# Tests class Args
def test_max_len_with_mixed_types():
    dic = {
        "latitude": np.array([45.0, 46.0]),
        "longitude": 10.0,
        "altitude": np.array([20.0, 25.0, 30.0]),
    }
    args = Parameters(dic)

    result = args.get_number_of_computations()

    assert result == 3


def test_max_len_with_ndarray():
    dic = {
        "latitude": np.array([45.0, 46.0]),
        "longitude": np.array([10.0, 11.0]),
    }
    args = Parameters(dic)

    result = args.get_number_of_computations()

    assert result == 2


def test_max_len_with_scalar():
    dic = {"latitude": 45.0, "longitude": 10.0}
    args = Parameters(dic)

    result = args.get_number_of_computations()

    assert result == 1


def test_max_len_with_empty_dict():
    args = Parameters({})

    result = args.get_number_of_computations()

    assert result == 1


def test_max_len_with_varied_lengths():
    dic = {
        "latitude": np.array([45.0, 46.0]),
        "longitude": np.array([10.0]),
        "altitude": np.array([20.0, 25.0, 30.0]),
    }
    args = Parameters(dic)

    result = args.get_number_of_computations()

    assert result == 3


def test_extend_to_max_len_with_nd_array():
    dic = {"latitude": np.array([45.0, 46.0]), "longitude": 10.0}
    args = Parameters(dic)

    args.extend()

    assert isinstance(args.latitude, ndarray)
    assert isinstance(args.longitude, ndarray)
    assert len(args.latitude) == 2
    assert len(args.longitude) == 2
    np.testing.assert_array_equal(args.latitude, np.array([45.0, 46.0]))
    np.testing.assert_array_equal(args.longitude, np.array([10.0, 10.0]))


def test_extend_to_max_len_with_scalar():
    dic = {"latitude": 45.0, "longitude": 10.0}
    args = Parameters(dic)

    args.extend()

    assert isinstance(args.latitude, ndarray)
    assert isinstance(args.longitude, ndarray)
    assert len(args.latitude) == 1
    assert len(args.longitude) == 1
    np.testing.assert_array_equal(args.latitude, np.array([45.0]))
    np.testing.assert_array_equal(args.longitude, np.array([10.0]))


def test_extend_to_max_len_with_mixed_types():
    dic = {
        "latitude": np.array([45.0, 46.0]),
        "longitude": 10.0,
        "altitude": 20.0,
    }
    args = Parameters(dic)

    args.extend()

    assert isinstance(args.latitude, ndarray)
    assert isinstance(args.longitude, ndarray)
    assert isinstance(args.altitude, ndarray)
    assert len(args.latitude) == 2
    assert len(args.longitude) == 2
    assert len(args.altitude) == 2
    np.testing.assert_array_equal(args.latitude, np.array([45.0, 46.0]))
    np.testing.assert_array_equal(args.longitude, np.array([10.0, 10.0]))
    np.testing.assert_array_equal(args.altitude, np.array([20.0, 20.0]))


def test_extend_to_max_len_with_empty_dict():
    args = Parameters({})

    args.extend()

    for key in args.keys():
        assert isinstance(args[key], (float, int, ndarray, datetime))
        if isinstance(args[key], ndarray):
            assert len(args[key]) == 1


def test_compress_with_unique_values():
    dic = {
        "latitude": np.array([45.0, 45.0]),
        "longitude": np.array([10.0, 10.0]),
    }
    args = Parameters(dic)

    args.compress()

    assert isinstance(args.latitude, float)
    assert args.latitude == 45.0
    assert isinstance(args.longitude, float)
    assert args.longitude == 10.0


def test_compress_with_non_unique_values():
    dic = {
        "latitude": np.array([45.0, 46.0]),
        "longitude": np.array([10.0, 11.0]),
    }
    args = Parameters(dic)

    args.compress()

    assert isinstance(args.latitude, ndarray)
    assert isinstance(args.longitude, ndarray)
    np.testing.assert_array_equal(args.latitude, np.array([45.0, 46.0]))
    np.testing.assert_array_equal(args.longitude, np.array([10.0, 11.0]))


def test_compress_with_mixed_values():
    dic = {
        "latitude": np.array([45.0, 45.0]),
        "longitude": np.array([10.0, 11.0]),
        "altitude": 20.0,
    }
    args = Parameters(dic)

    args.compress()

    assert isinstance(args.latitude, float)
    assert args.latitude == 45.0
    assert isinstance(args.longitude, ndarray)
    assert isinstance(args.altitude, float)
    np.testing.assert_array_equal(args.longitude, np.array([10.0, 11.0]))
    assert args.altitude == 20.0


def test_compress_with_empty_dict():
    args = Parameters({})
    args.compress()
    for key in args.keys():
        assert isinstance(args[key], (float, np.int64, ndarray, datetime))


# Tests Fonctions Base
def test_reshape_on_one_row_array():
    array = np.array([1.0, 2.0, 3.0])
    nb_row = 3
    nb_columns = 1
    expected = np.array([[1.0], [2.0], [3.0]])

    result = reshape(array, nb_row, nb_columns)

    np.testing.assert_array_equal(result, expected)


def test_reshape_1d_to_2d_column_stack():
    array = np.array([1.0, 2.0, 3.0])
    nb_row = 3
    nb_columns = 1
    expected = np.array([[1.0], [2.0], [3.0]])

    result = reshape(array, nb_row, nb_columns)

    np.testing.assert_array_equal(result, expected)


def test_reshape_1d_to_2d_row_stack():
    array = np.array([1.0, 2.0, 3.0])
    nb_row = 1
    nb_columns = 3
    expected = np.array([[1.0, 2.0, 3.0]])

    result = reshape(array, nb_row, nb_columns)

    np.testing.assert_array_equal(result, expected)


def test_reshape_2d_to_2d():
    array = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    nb_row = 2
    nb_columns = 3
    expected = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])

    result = reshape(array, nb_row, nb_columns)

    np.testing.assert_array_equal(result, expected)


def test_reshape_scalar_to_2d():
    scalar = 1.0
    nb_row = 2
    nb_columns = 3
    expected = np.array([[1.0, 1.0, 1.0], [1.0, 1.0, 1.0]])

    result = reshape(scalar, nb_row, nb_columns)

    np.testing.assert_array_equal(result, expected)


def test_reshape_invalid_shape():
    array = np.array(0)  # ([1.0, 2.0, 3.0])
    nb_row = 2
    nb_columns = 2
    expected = np.array([[0, 0], [0, 0]])

    result = reshape(array, nb_row, nb_columns)

    np.testing.assert_array_equal(result, expected)


def test_set_dates_single_day():
    datetime_utc = datetime(2000, 1, 1, 0, tzinfo=timezone.utc)
    offset = np.array([0, 3600, 7200])
    n = 1

    result = _set_dates(datetime_utc, offset, n)

    assert result.shape == (3, 1)
    assert result[0, 0] == datetime(2000, 1, 1, 0, tzinfo=timezone.utc)
    assert result[1, 0] == datetime(2000, 1, 1, 1, tzinfo=timezone.utc)
    assert result[2, 0] == datetime(2000, 1, 1, 2, tzinfo=timezone.utc)


def test_set_dates_multiple_days():
    datetime_utc = datetime(2000, 1, 1, 23, tzinfo=timezone.utc)
    offset = np.array([0, 3600, 7200])
    n = 1

    result = _set_dates(datetime_utc, offset, n)

    assert result.shape == (3, 1)
    assert result[0, 0] == datetime(2000, 1, 1, 23, tzinfo=timezone.utc)
    assert result[1, 0] == datetime(2000, 1, 2, 0, tzinfo=timezone.utc)
    assert result[2, 0] == datetime(2000, 1, 2, 1, tzinfo=timezone.utc)


def test_set_dates_multiple_months():
    datetime_utc = datetime(2000, 12, 31, 23, tzinfo=timezone.utc)
    offset = np.array([0, 3600, 7200])
    n = 1

    result = _set_dates(datetime_utc, offset, n)

    assert result.shape == (3, 1)
    assert result[0, 0] == datetime(2000, 12, 31, 23, tzinfo=timezone.utc)
    assert result[1, 0] == datetime(2001, 1, 1, 0, tzinfo=timezone.utc)
    assert result[2, 0] == datetime(2001, 1, 1, 1, tzinfo=timezone.utc)


def test_set_dates_multiple_inputs():
    datetime_utc = [
        datetime(2000, 1, 1, 0, tzinfo=timezone.utc),
        datetime(2000, 2, 2, 12, tzinfo=timezone.utc),
    ]
    offset = np.array([0, 3600, 7200])
    n = 2

    result = _set_dates(datetime_utc, offset, n)

    assert result.shape == (3, 2)

    assert result[0, 0] == datetime(2000, 1, 1, 0, tzinfo=timezone.utc)
    assert result[1, 0] == datetime(2000, 1, 1, 1, tzinfo=timezone.utc)
    assert result[2, 0] == datetime(2000, 1, 1, 2, tzinfo=timezone.utc)

    assert result[0, 1] == datetime(2000, 2, 2, 12, tzinfo=timezone.utc)
    assert result[1, 1] == datetime(2000, 2, 2, 13, tzinfo=timezone.utc)
    assert result[2, 1] == datetime(2000, 2, 2, 14, tzinfo=timezone.utc)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np

from thermohl.power.rte import Air


def test_volumic_mass_scalar():
    air_temperature = 25.0
    altitude = 1000.0
    expected = (1.293 - 1.525e-04 * altitude + 6.379e-09 * altitude**2) / (
        1.0 + 0.00367 * air_temperature
    )

    result = Air.volumic_mass(air_temperature, altitude)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_volumic_mass_array():
    air_temperature = np.array([25.0, 30.0])
    altitude = np.array([1000.0, 2000.0])
    expected = (1.293 - 1.525e-04 * altitude + 6.379e-09 * altitude**2) / (
        1.0 + 0.00367 * air_temperature
    )

    result = Air.volumic_mass(air_temperature, altitude)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_volumic_mass_default_altitude():
    air_temperature = 25.0
    expected = (1.293 - 1.525e-04 * 0.0 + 6.379e-09 * 0.0**2) / (
        1.0 + 0.00367 * air_temperature
    )

    result = Air.volumic_mass(air_temperature)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_volumic_mass_array_default_altitude():
    air_temperature = np.array([25.0, 30.0])
    expected = (1.293 - 1.525e-04 * 0.0 + 6.379e-09 * 0.0**2) / (
        1.0 + 0.00367 * air_temperature
    )

    result = Air.volumic_mass(air_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_dynamic_viscosity_scalar():
    air_temperature = 25.0
    expected = (1.458e-06 * (air_temperature + 273.0) ** 1.5) / (
        air_temperature + 383.4
    )

    result = Air.dynamic_viscosity(air_temperature)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_dynamic_viscosity_array():
    air_temperature = np.array([25.0, 30.0])
    expected = (1.458e-06 * (air_temperature + 273.0) ** 1.5) / (
        air_temperature + 383.4
    )

    result = Air.dynamic_viscosity(air_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_thermal_conductivity_scalar():
    air_temperature = 25.0
    expected = 2.424e-02 + 7.477e-05 * air_temperature - 4.407e-09 * air_temperature**2

    result = Air.thermal_conductivity(air_temperature)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_thermal_conductivity_array():
    air_temperature = np.array([25.0, 30.0])
    expected = 2.424e-02 + 7.477e-05 * air_temperature - 4.407e-09 * air_temperature**2

    result = Air.thermal_conductivity(air_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

# mypy: ignore-errors
"""Rain cooling power term.

Based on *Modelling precipitation cooling of overhead conductors*, Pytlak et
al., 2011 (https://doi.org/10.1016/j.epsr.2011.06.004) and *Dynamic thermal
rating of power lines in raining conditions-model and measurements*, Maksic et
al., 2016 (https://doi.org/10.1109/PESGM.2016.7741611).
"""

import numpy as np

from thermohl import air
from thermohl import thermodynamics
from thermohl.power import olla, ieee
from thermohl.utils import PowerTerm


class PrecipitationCooling(PowerTerm):
    """Precipitation cooling term."""

    def value(
        self,
        conductor_temperature,
        altitude,
        ambient_temp,
        wind_speed,
        wind_azimuth,
        outer_diameter,
        air_pressure,
        relative_humidity,
        rain_rate,
        **kwargs,
    ):
        # maksic = (
        #     PrecipitationCooling._evap(
        #         conductor_temperature,
        #         altitude,
        #         ambient_temp,
        #         wind_speed,
        #         wind_azimuth,
        #         outer_diameter,
        #         air_pressure,
        #         relative_humidity,
        #         rain_rate,
        #     )
        #     + PrecipitationCooling._imp(
        #         conductor_temperature,
        #         ambient_temp,
        #         wind_speed,
        #         outer_diameter,
        #         rain_rate,
        #     )
        # )
        pytlak = PrecipitationCooling._evap(
            conductor_temperature,
            altitude,
            ambient_temp,
            wind_speed,
            wind_azimuth,
            outer_diameter,
            air_pressure,
            relative_humidity,
            rain_rate,
        )
        return pytlak

    @staticmethod
    def _ma(
        ambient_temp,
        wind_speed,
        outer_diameter,
        rain_rate,
        snow_rate,
    ):
        # ! -- precipitation_rate and ps in m.s**-1
        water_density = thermodynamics.Water.volumic_mass(ambient_temp)
        rain_mass_flux = np.sqrt(
            (rain_rate * water_density) ** 2
            + (wind_speed * 23.589 * rain_rate**0.8460) ** 2
        )
        snow_mass_flux = np.sqrt(
            (snow_rate * water_density) ** 2
            + (wind_speed * 142.88 * snow_rate**0.9165) ** 2
        )
        return outer_diameter * (rain_mass_flux + snow_mass_flux)

    @staticmethod
    def _me(
        conductor_temperature,
        altitude,
        ambient_temp,
        wind_speed,
        wind_azimuth,
        outer_diameter,
        air_pressure,
        relative_humidity,
    ):
        film_temperature = 0.5 * (conductor_temperature + ambient_temp)
        temperature_delta = conductor_temperature - ambient_temp
        air_density = air.IEEE.volumic_mass(film_temperature, altitude)
        convective_cooling = ieee.ConvectiveCooling._value_forced(
            film_temperature,
            temperature_delta,
            air_density,
            wind_speed,
            outer_diameter,
            wind_azimuth,
        )
        heat_transfer_coeff = convective_cooling / (
            np.pi * outer_diameter * temperature_delta
        )
        mass_ratio = 18.015 / 28.9647
        wetted_perimeter = np.pi * outer_diameter
        air_heat_capacity = thermodynamics.Air.heat_capacity(
            temperature=olla.air.kelvin(film_temperature)
        )
        vapor_pressure_conductor = thermodynamics.Water.vapor_pressure(
            temperature=olla.air.kelvin(conductor_temperature)
        )
        vapor_pressure_ambient = thermodynamics.Water.vapor_pressure(
            temperature=olla.air.kelvin(ambient_temp)
        )
        evaporation_mass_flux = (
            wetted_perimeter
            * heat_transfer_coeff
            * mass_ratio
            * (vapor_pressure_conductor - relative_humidity * vapor_pressure_ambient)
            / (air_heat_capacity * air_pressure)
        )
        return np.where(
            ~np.isclose(temperature_delta, 0.0, atol=0.0005),
            evaporation_mass_flux,
            np.zeros_like(conductor_temperature),
        )

    @staticmethod
    def _mass_flux(
        conductor_temperature,
        altitude,
        ambient_temp,
        wind_speed,
        wind_azimuth,
        outer_diameter,
        air_pressure,
        relative_humidity,
        rain_rate,
        snow_rate,
    ):
        return np.minimum(
            PrecipitationCooling._ma(
                ambient_temp,
                wind_speed,
                outer_diameter,
                rain_rate,
                snow_rate,
            ),
            PrecipitationCooling._me(
                conductor_temperature,
                altitude,
                ambient_temp,
                wind_speed,
                wind_azimuth,
                outer_diameter,
                air_pressure,
                relative_humidity,
            ),
        )

    @staticmethod
    def _evap(
        conductor_temperature,
        altitude,
        ambient_temp,
        wind_speed,
        wind_azimuth,
        outer_diameter,
        air_pressure,
        relative_humidity,
        rain_rate,
        snow_rate=0.0,
    ):
        mass_flux = PrecipitationCooling._mass_flux(
            conductor_temperature,
            altitude,
            ambient_temp,
            wind_speed,
            wind_azimuth,
            outer_diameter,
            air_pressure,
            relative_humidity,
            rain_rate,
            snow_rate,
        )
        latent_heat_vap = thermodynamics.Water.heat_of_vaporization()
        water_heat_capacity = thermodynamics.Water.heat_capacity(
            temperature=olla.air.kelvin(conductor_temperature)
        )
        boiling_temp = thermodynamics.Water.boiling_point(pressure=air_pressure)
        boiling_temp_c = air.celsius(boiling_temp)
        effective_temp = np.minimum(conductor_temperature, boiling_temp_c)
        latent_heat_fusion = thermodynamics.Ice.heat_of_fusion()
        ice_heat_capacity = thermodynamics.Ice.heat_capacity()
        rain_cooling = mass_flux * (
            latent_heat_vap + water_heat_capacity * (effective_temp - ambient_temp)
        )
        snow_cooling = mass_flux * (
            latent_heat_vap
            + water_heat_capacity * conductor_temperature
            + latent_heat_fusion * ice_heat_capacity * ambient_temp
        )
        return np.maximum(
            rain_cooling + 0.0 * snow_cooling, np.zeros_like(conductor_temperature)
        )

    @staticmethod
    def _imp(
        conductor_temperature,
        ambient_temp,
        wind_speed,
        outer_diameter,
        rain_rate,
        snow_rate=0.0,
    ):
        water_heat_capacity = thermodynamics.Water.heat_capacity(
            temperature=olla.air.kelvin(conductor_temperature)
        )
        return (
            0.71
            * water_heat_capacity
            * (conductor_temperature - ambient_temp)
            * PrecipitationCooling._ma(
                ambient_temp,
                wind_speed,
                outer_diameter,
                rain_rate,
                snow_rate,
            )
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Tuple, Type, Optional, Any, Callable, Final, Dict

import numpy as np

from thermohl import floatArrayLike, floatArray, intArray
from thermohl.power import PowerTerm
from thermohl.solver.entities import TargetType, CableLocationListLike
from thermohl.solver.slv3t import Solver3T


class Solver3TL(Solver3T):
    def __init__(
        self,
        dic: Optional[dict[str, Any]] = None,
        joule: Type[PowerTerm] = PowerTerm,
        solar: Type[PowerTerm] = PowerTerm,
        convective: Type[PowerTerm] = PowerTerm,
        radiative: Type[PowerTerm] = PowerTerm,
        precipitation: Type[PowerTerm] = PowerTerm,
    ):
        super().__init__(dic, joule, solar, convective, radiative, precipitation)
        self.update()

    def _morgan_coefficients(self) -> Tuple[floatArray, intArray]:
        """
        Calculate coefficients for heat flux between surface and core in steady state.

        Returns:
            Tuple[numpy.ndarray[float], numpy.ndarray[float], numpy.ndarray[float], numpy.ndarray[int]]
                - heat_flux_coefficients : numpy.ndarray[float]
                    Coefficient array for heat flux.
                - indices_non_zero_diameter : numpy.ndarray[int]
                    Indices where core diameter is greater than 0.
                    When conductors are uniform, core diameter is equal to 0.0.
                    When conductors are bimetallic, core diameter is greater than 0.0.
        """
        UNIFORM_CONDUCTOR_COEFFICIENT: Final[float] = 1 / 13
        BIMETALLIC_CONDUCTOR_COEFFICIENT: Final[float] = 1 / 21

        core_diameter_array = self.args.core_diameter * np.ones(
            (self.args.get_number_of_computations(),)
        )
        indices_non_zero_diameter = np.nonzero(core_diameter_array > 0.0)[0]
        heat_flux_coefficients = UNIFORM_CONDUCTOR_COEFFICIENT * np.ones_like(
            core_diameter_array
        )
        heat_flux_coefficients[indices_non_zero_diameter] = (
            BIMETALLIC_CONDUCTOR_COEFFICIENT
        )
        return heat_flux_coefficients, indices_non_zero_diameter

    def average(self, surface_temperature, core_temperature):
        """
        Compute average temperature given surface and core temperature.

        Unlike Solver3T, always use a regular mean even for non-homogeneous
        conductors.

        Args:
            surface_temperature (numpy.ndarray): Array of surface temperatures.
            core_temperature (numpy.ndarray): Array of core temperatures.
        """
        return 0.5 * (surface_temperature + core_temperature)

    def morgan(
        self, surface_temperature: floatArray, core_temperature: floatArray
    ) -> floatArray:
        """
        Computes the Morgan function for given temperature arrays.

        Args:
            surface_temperature (numpy.ndarray): Array of surface temperatures.
            core_temperature (numpy.ndarray): Array of core temperatures.

        Returns:
            numpy.ndarray: Resulting array after applying the Morgan function.
        """
        morgan_coefficient = self.morgan_coefficients[0]
        return (
            core_temperature - surface_temperature
        ) - morgan_coefficient * self.joule(surface_temperature, core_temperature)

    def _steady_intensity_header(
        self, T: floatArrayLike, target: CableLocationListLike
    ) -> Tuple[np.ndarray, Callable]:
        """Format input for ampacity solver."""

        max_len = self.args.get_number_of_computations()
        Tmax = T * np.ones(max_len)
        target_ = self._check_target(target, self.args.core_diameter, max_len)

        # pre-compute indexes
        surface_indices = np.nonzero(target_ == TargetType.SURFACE)[0]
        average_indices = np.nonzero(target_ == TargetType.AVERAGE)[0]
        core_indices = np.nonzero(target_ == TargetType.CORE)[0]

        def newtheader(
            transit: floatArray, tg: floatArray
        ) -> Tuple[floatArray, floatArray]:
            self.args.transit = transit
            self.joule_heating.__init__(**self.args.__dict__)
            surface_temperature = np.ones_like(tg) * np.nan
            core_temperature = np.ones_like(tg) * np.nan

            surface_temperature[surface_indices] = Tmax[surface_indices]
            core_temperature[surface_indices] = tg[surface_indices]

            surface_temperature[average_indices] = tg[average_indices]
            core_temperature[average_indices] = (
                2 * Tmax[average_indices] - surface_temperature[average_indices]
            )

            core_temperature[core_indices] = Tmax[core_indices]
            surface_temperature[core_indices] = tg[core_indices]

            return surface_temperature, core_temperature

        return Tmax, newtheader

    def _morgan_transient(self):
        """Morgan coefficients for transient temperature."""
        morgan_coeff_1, _ = self.morgan_coefficients
        morgan_coeff_2 = 0.5 * np.ones_like(morgan_coeff_1)
        return morgan_coeff_1, morgan_coeff_2

    def transient_temperature_legacy(
        self,
        offset: floatArray = np.array([]),
        surface_temperature_0: Optional[floatArrayLike] = None,
        core_temperature_0: Optional[floatArrayLike] = None,
        time_constant: float = 600.0,
        return_power: bool = False,
    ) -> Dict[str, Any]:
        """
        Compute transient-state temperature with legacy method.

        Args:
            offset (numpy.ndarray): A 1D array with times (in seconds) when the temperature needs to be
                computed. The array must contain increasing values (undefined behaviour otherwise).
            surface_temperature_0 (float): Initial surface temperature. If set to None, the ambient temperature from
                internal dict will be used. The default is None.
            core_temperature_0 (float): Initial core temperature. If set to None, the ambient temperature from
                internal dict will be used. The default is None.
            time_constant (float): A time-constant to add some inertia. The default is 600.
            return_power (bool, optional): Return power term values. The default is False.

        Returns:
            Dict[str, Any]: A dictionary with temperature and other results (depending on inputs)
                in the keys.

        """

        # get sizes (n for input dict entries, N for offsets)
        n = self.args.get_number_of_computations()
        N = len(offset)
        if N < 2:
            raise ValueError()

        # get initial temperature
        surface_temperature_0 = (
            surface_temperature_0
            if surface_temperature_0 is not None
            else self.args.ambient_temperature
        )
        core_temperature_0 = (
            core_temperature_0
            if core_temperature_0 is not None
            else 1.0 + surface_temperature_0
        )

        # inverse of m*C : shortcuts for time-loop
        imc = 1.0 / (self.args.linear_mass * self.args.heat_capacity)

        # define temperature data 2D arrays
        surface_temperature = np.zeros((N, n))
        average_temperature = np.zeros((N, n))
        core_temperature = np.zeros((N, n))
        temperature_difference = np.zeros((N, n))
        # set initial values in first row
        surface_temperature[0, :] = surface_temperature_0
        core_temperature[0, :] = core_temperature_0
        average_temperature[0, :] = self.average(
            surface_temperature[0, :], core_temperature[0, :]
        )
        temperature_difference[0, :] = (
            core_temperature[0, :] - surface_temperature[0, :]
        )

        # compute transient temperatures for each row after the first.
        for i in range(1, len(offset)):
            balance = self.balance(
                surface_temperature[i - 1, :], core_temperature[i - 1, :]
            )
            time_difference = offset[i] - offset[i - 1]
            average_temperature[i, :] = (
                average_temperature[i - 1, :] + time_difference * imc * balance
            )
            temperature_difference[i, :] = temperature_difference[i - 1, :] * (
                1.0 - time_difference / time_constant
            ) + (
                time_difference
                / time_constant
                * self.morgan_coefficients[0]
                * self.joule_heating.value(average_temperature[i, :])
            )
            core_temperature[i, :] = (
                average_temperature[i, :] + 0.5 * temperature_difference[i, :]
            )
            surface_temperature[i, :] = (
                average_temperature[i, :] - 0.5 * temperature_difference[i, :]
            )

        return self._transient_temperature_results(
            offset,
            surface_temperature,
            average_temperature,
            core_temperature,
            return_power,
            n,
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np

from thermohl.power import PowerTerm


# Tests for PowerTerm
def test_power_term_value_with_scalar():
    power_term = PowerTerm()
    conductor_temperature = 25.0

    result = power_term.value(conductor_temperature)

    assert result == 0.0


def test_power_term_value_with_array():
    power_term = PowerTerm()
    conductor_temperature = np.array([25.0, 30.0, 35.0])
    expected = np.array([0.0, 0.0, 0.0])

    result = power_term.value(conductor_temperature)

    np.testing.assert_array_equal(result, expected)


def test_power_term_derivative_with_scalar():
    power_term = PowerTerm()
    temperature = 25.0

    result = power_term.derivative(temperature)

    assert result == 0.0


def test_power_term_derivative_with_array():
    power_term = PowerTerm()
    temperatures = np.array([25.0, 30.0, 35.0])
    expected = np.array([0.0, 0.0, 0.0])

    result = power_term.derivative(temperatures)

    np.testing.assert_array_equal(result, expected)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any

import numpy as np

from thermohl import floatArrayLike
from thermohl.power import PowerTerm


class JouleHeating(PowerTerm):
    """Joule heating term."""

    @staticmethod
    def _c(
        temp_low: floatArrayLike,
        temp_high: floatArrayLike,
        linear_resistance_temp_low: floatArrayLike,
        linear_resistance_temp_high: floatArrayLike,
    ) -> floatArrayLike:
        return (linear_resistance_temp_high - linear_resistance_temp_low) / (
            temp_high - temp_low
        )

    def __init__(
        self,
        transit: floatArrayLike,
        temp_low: floatArrayLike,
        temp_high: floatArrayLike,
        linear_resistance_temp_low: floatArrayLike,
        linear_resistance_temp_high: floatArrayLike,
        **kwargs: Any,
    ):
        r"""Init with args.

        If more than one input are numpy arrays, they should have the same size.

        Args:
            transit (float | numpy.ndarray): Transit intensity (A).
            temp_low (float | numpy.ndarray): Temperature for linear_resistance_temp_low measurement (°C).
            temp_high (float | numpy.ndarray): Temperature for linear_resistance_temp_high measurement (°C).
            linear_resistance_temp_low (float | numpy.ndarray): Electric resistance per unit length at temp_low (Ω·m⁻¹).
            linear_resistance_temp_high (float | numpy.ndarray): Electric resistance per unit length at temp_high (Ω·m⁻¹).

        """
        self.temp_low = temp_low
        self.temp_high = temp_high
        self.linear_resistance_temp_low = linear_resistance_temp_low
        self.linear_resistance_temp_high = linear_resistance_temp_high
        self.transit = transit
        self.temp_coeff_linear = JouleHeating._c(
            temp_low,
            temp_high,
            linear_resistance_temp_low,
            linear_resistance_temp_high,
        )

    def _rdc(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        return self.linear_resistance_temp_low + self.temp_coeff_linear * (
            conductor_temperature - self.temp_low
        )

    def value(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        r"""Compute joule heating.

        Args:
            conductor_temperature (float | numpy.ndarray): Conductor temperature (°C).

        Returns:
            float | numpy.ndarray: Power term value (W·m⁻¹).

        """
        return self._rdc(conductor_temperature) * self.transit**2

    def derivative(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        r"""Compute joule heating derivative.

        If more than one input are numpy arrays, they should have the same size.

        Args:
            conductor_temperature (float | numpy.ndarray): Conductor temperature (°C).

        Returns:
            float | numpy.ndarray: Power term derivative (W·m⁻¹·K⁻¹).

        """
        return (
            self.temp_coeff_linear
            * self.transit**2
            * np.ones_like(conductor_temperature)
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timezone
from typing import Optional, Any, KeysView, Iterable
import numpy as np


class DEFAULT_PARAMETERS:
    tmin = -99.0
    tmax = +999.0
    tol = 1.0e-09
    maxiter = 64
    imin = 0.0
    imax = 9999.0


class Parameters:
    """Object to store Solver args in a dict-like manner."""

    def __init__(self, input_dict: Optional[dict[str, Any]] = None):
        # add default values
        self._set_default_values()
        # use values from input dict
        if input_dict is None:
            return
        arg_keys = self.keys()
        for input_key, input_value in input_dict.items():
            if input_key in arg_keys and input_value is not None:
                self[input_key] = input_value

    def _set_default_values(self) -> None:
        """Set default values."""
        # position
        self.latitude = 45.0  # latitude (deg)
        self.longitude = 0.0  # longitude (deg)
        self.altitude = 0.0  # altitude (m)
        self.cable_azimuth = 0.0  # cable_azimuth (deg)
        self.datetime_utc = datetime(2000, 3, 21, 12, tzinfo=timezone.utc)

        # weather and mesurement
        self.measured_solar_irradiance = np.nan  # solar irradiance
        self.ambient_temperature = 15.0  # ambient temperature (C)
        self.ambient_pressure = 1.0e05  # ambient pressure (Pa)
        self.relative_humidity = 0.8  # relative humidity (none, in [0, 1])
        self.precipitation_rate = 0.0  # rain precipitation rate (m.s**-1)
        self.wind_speed = 0.0  # wind speed (m.s**-1)
        self.wind_azimuth = 90.0  # wind_azimuth (deg, regarding north)
        self.nebulosity = 0  # nebulosity (1)
        self.albedo = 0.15  # albedo (1)
        # coefficient for air pollution from 0 (clean) to 1 (polluted)
        self.turbidity = 0.1
        self.transit = 100.0  # transit intensity (A)

        # conductor
        self.linear_mass = 1.5  # mass per unit length (kg.m**-1)
        self.core_diameter = 1.9e-02  # core diameter (m)
        self.outer_diameter = 3.0e-02  # external (global) diameter (m)
        self.core_area = 2.84e-04  # core section (m**2)
        self.outer_area = 7.07e-04  # external (global) section (m**2)
        self.roughness_ratio = 4.0e-02  # roughness (1)
        # radial thermal conductivity (W.m**-1.K**-1)
        self.radial_thermal_conductivity = 1.0
        self.heat_capacity = 500.0  # specific heat capacity (J.kg**-1.K**-1)
        self.solar_absorptivity = 0.5  # solar absorption (1)
        self.emissivity = 0.5  # emissivity (1)
        # electric resistance per unit length (DC) at 20°C (Ohm.m**-1)
        self.linear_resistance_dc_20c = 2.5e-05
        self.magnetic_coeff = 1.006  # coefficient for magnetic effects (1)
        self.magnetic_coeff_per_a = 0.016  # coefficient for magnetic effects (A**-1)
        # linear resistance augmentation with temperature (K**-1)
        self.temperature_coeff_linear = 3.8e-03
        # quadratic resistance augmentation with temperature (K**-2)
        self.temperature_coeff_quadratic = 8.0e-07
        # electric resistance per unit length (DC) at temp_high (Ohm.m**-1)
        self.linear_resistance_temp_high = 3.05e-05
        # electric resistance per unit length (DC) at temp_low (Ohm.m**-1)
        self.linear_resistance_temp_low = 2.66e-05
        # temperature for linear_resistance_temp_high measurement (°C)
        self.temp_high = 60.0
        # temperature for linear_resistance_temp_low measurement (°C)
        self.temp_low = 20.0

    def keys(self) -> KeysView[str]:
        """Get list of members as dict keys."""
        return self.__dict__.keys()

    def __getitem__(self, key: str) -> Any:
        return self.__dict__[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.__dict__[key] = value

    def get_number_of_computations(self) -> int:
        return max(
            (len(self[k]) for k in self.keys() if isinstance(self[k], Iterable)),
            default=1,
        )

    def extend(self) -> None:
        """
        Extend all compressed elements in the Args dictionary to the right length, ie the number of computations.
        If the element is a list, it already has the right length.
        If the element is a scalar, it is replaced with a list of the right length filled with the scalar value.
        """
        number_of_computations = self.get_number_of_computations()
        for key in self.keys():
            if not isinstance(self[key], Iterable):
                self[key] = np.array(number_of_computations * [self[key]])

    def compress(self) -> None:
        """
        Compress the elements in the Args dictionary by replacing lists containing a unique value
        with this value.
        """
        for key in self.keys():
            u = np.unique(self[key])
            if len(u) == 1:
                self[key] = u[0]



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any

from thermohl import floatArrayLike
from thermohl.power import PowerTerm


class RadiativeCooling(PowerTerm):
    """Power term for radiative cooling.

    Very similar to RadiativeCooling. Difference are in the
    Stefan-Boltzman constant value and the celsius-kelvin conversion.
    """

    def __init__(
        self,
        ambient_temperature: floatArrayLike,
        outer_diameter: floatArrayLike,
        emissivity: floatArrayLike,
        **kwargs: Any,
    ):
        r"""Init with args.

        Args:
            ambient_temperature (float | numpy.ndarray): Ambient temperature (°C).
            outer_diameter (float | numpy.ndarray): External diameter (m).
            emissivity (float | numpy.ndarray): Emissivity (—).

        """
        self.ambient_temp = ambient_temperature
        self.outer_diameter = outer_diameter
        self.emissivity = emissivity

    def value(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        r"""Compute radiative cooling using the Stefan-Boltzmann law.

        Args:
            conductor_temperature (float | numpy.ndarray): Conductor temperature (°C).

        Returns:
            float | numpy.ndarray: Power term value (W·m⁻¹).

        """
        return (
            17.8
            * self.emissivity
            * self.outer_diameter
            * (
                ((conductor_temperature + 273.0) / 100.0) ** 4
                - ((self.ambient_temp + 273.0) / 100.0) ** 4
            )
        )

    def derivative(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        r"""Analytical derivative of value method.

        Args:
            conductor_temperature (float | numpy.ndarray): Conductor temperature (K).

        Returns:
            float | numpy.ndarray: Power term derivative (W·m⁻¹·K⁻¹).

        """
        return (
            4.0
            * 1.78e-07
            * self.emissivity
            * self.outer_diameter
            * conductor_temperature**3
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import pytest
import numpy as np

from thermohl import solver
from thermohl.power.convective_cooling import compute_wind_attack_angle
from thermohl.power.ieee import ConvectiveCooling


def set_default_values_scalar():
    dic = solver.default_values()
    dic["wind_speed"] = 0.61
    dic["wind_azimuth"] = 0.0
    dic["emissivity"] = 0.8
    dic["solar_absorptivity"] = 0.8
    dic["ambient_temperature"] = 40.0
    dic["temp_high"] = 75.0
    dic["temp_low"] = 25.0
    dic["linear_resistance_temp_high"] = 8.688e-05
    dic["linear_resistance_temp_low"] = 7.283e-05
    dic["cable_azimuth"] = 90.0
    dic["latitude"] = 30.0
    dic["turbidity"] = 0.0
    dic["altitude"] = 0.0
    dic["outer_diameter"] = 28.14 * 1.0e-03
    dic["core_diameter"] = 10.4 * 1.0e-03
    dic["month"] = 6
    dic["day"] = 10
    dic["hour"] = 11.0
    return dic


def set_default_values_array():
    dic = solver.default_values()
    dic["wind_speed"] = np.array([0.61, 0.83])
    dic["wind_azimuth"] = np.array([0.0, 42.1])
    dic["emissivity"] = np.array([0.8, 0.9])
    dic["solar_absorptivity"] = np.array([0.8, 0.9])
    dic["ambient_temperature"] = np.array([40.0, 32])
    dic["temp_high"] = np.array([75.0, 70.0])
    dic["temp_low"] = np.array([25.0, 20])
    dic["linear_resistance_temp_high"] = np.array([8.688e-05, 8.688e-05])
    dic["linear_resistance_temp_low"] = np.array([7.283e-05, 7.283e-05])
    dic["cable_azimuth"] = np.array([90.0, 90.0])
    dic["latitude"] = np.array([30.0, 30.0])
    dic["turbidity"] = np.array([0.0, 0.0])
    dic["altitude"] = np.array([0.0, 0.0])
    dic["outer_diameter"] = np.array([28.14 * 1.0e-03, 28.14 * 1.0e-03])
    dic["core_diameter"] = np.array([10.4 * 1.0e-03, 10.4 * 1.0e-03])
    dic["month"] = np.array([6, 3])
    dic["day"] = np.array([10, 13])
    dic["hour"] = np.array([11.0, 8.0])
    return dic


convective_cooling_instances = [
    (ConvectiveCooling(**set_default_values_array()), np.array([80.9093, 82.3549])),
    (
        ConvectiveCooling(
            **set_default_values_scalar(),
        ),
        80.9,
    ),
]


@pytest.mark.parametrize(
    "convective_cooling, expected",
    convective_cooling_instances,
    ids=[
        "ConvectiveCooling with arrays",
        "ConvectiveCooling with scalars",
    ],
)
def test_value_forced_scalar(convective_cooling, expected):
    """Test the _value_forced method of ConvectiveCooling."""
    Tf = 70.0
    Td = 60.0
    vm = 1.0

    result = convective_cooling._value_forced(Tf, Td, vm)

    assert np.allclose(result, expected, rtol=0.002)


@pytest.mark.parametrize(
    "convective_cooling, expected",
    convective_cooling_instances,
    ids=[
        "ConvectiveCooling with arrays",
        "ConvectiveCooling with scalars",
    ],
)
def test_value_forced_array(convective_cooling, expected):
    """Test the _value_forced method of ConvectiveCooling."""
    Tf = np.array([70.0, 70.0])
    Td = np.array([60.0, 60.0])
    vm = np.array([1.0, 1.0])

    result = convective_cooling._value_forced(Tf, Td, vm)

    assert np.allclose(result, expected, rtol=0.002)


@pytest.mark.parametrize(
    "convective_cooling, expected",
    convective_cooling_instances,
    ids=[
        "ConvectiveCooling with arrays",
        "ConvectiveCooling with scalars",
    ],
)
def test_value_natural_scalar(convective_cooling, expected):
    """Test the _value_natural method of ConvectiveCooling."""
    Td = 60.0
    vm = 1.0

    result = convective_cooling._value_natural(Td, vm)

    expected_result = (
        3.645
        * np.sqrt(vm)
        * convective_cooling.outer_diameter**0.75
        * np.sign(Td)
        * np.abs(Td) ** 1.25
    )
    assert np.allclose(result, expected_result, rtol=0.002)


@pytest.mark.parametrize(
    "convective_cooling, expected",
    convective_cooling_instances,
    ids=[
        "ConvectiveCooling with arrays",
        "ConvectiveCooling with scalars",
    ],
)
def test_value_natural_array(convective_cooling, expected):
    """Test the _value_natural method of ConvectiveCooling."""
    Td = np.array([60.0, 60.0])
    vm = np.array([1.0, 1.0])

    result = convective_cooling._value_natural(Td, vm)

    expected_result = (
        3.645
        * np.sqrt(vm)
        * convective_cooling.outer_diameter**0.75
        * np.sign(Td)
        * np.abs(Td) ** 1.25
    )
    assert np.allclose(result, expected_result, rtol=0.002)


@pytest.mark.parametrize(
    "convective_cooling, expected",
    convective_cooling_instances,
    ids=[
        "ConvectiveCooling with arrays",
        "ConvectiveCooling with scalars",
    ],
)
def test_convective_cooling_value_scalar(convective_cooling, expected):
    """Test the value method of ConvectiveCooling."""
    conductor_temperature = 100.0

    result = convective_cooling.value(conductor_temperature)

    expected_result = np.maximum(
        convective_cooling._value_forced(
            0.5 * (conductor_temperature + convective_cooling.ambient_temp),
            conductor_temperature - convective_cooling.ambient_temp,
            convective_cooling.air_density(
                0.5 * (conductor_temperature + convective_cooling.ambient_temp),
                convective_cooling.altitude,
            ),
        ),
        convective_cooling._value_natural(
            conductor_temperature - convective_cooling.ambient_temp,
            convective_cooling.air_density(
                0.5 * (conductor_temperature + convective_cooling.ambient_temp),
                convective_cooling.altitude,
            ),
        ),
    )
    assert np.allclose(result, expected_result, rtol=0.002)


@pytest.mark.parametrize(
    "convective_cooling, expected",
    convective_cooling_instances,
    ids=[
        "ConvectiveCooling with arrays",
        "ConvectiveCooling with scalars",
    ],
)
def test_convective_cooling_value_array(convective_cooling, expected):
    """Test the value method of ConvectiveCooling."""
    conductor_temperature = np.array([60.3, 100.0])

    result = convective_cooling.value(conductor_temperature)

    expected_result = np.maximum(
        convective_cooling._value_forced(
            0.5 * (conductor_temperature + convective_cooling.ambient_temp),
            conductor_temperature - convective_cooling.ambient_temp,
            convective_cooling.air_density(
                0.5 * (conductor_temperature + convective_cooling.ambient_temp),
                convective_cooling.altitude,
            ),
        ),
        convective_cooling._value_natural(
            conductor_temperature - convective_cooling.ambient_temp,
            convective_cooling.air_density(
                0.5 * (conductor_temperature + convective_cooling.ambient_temp),
                convective_cooling.altitude,
            ),
        ),
    )
    assert np.allclose(result, expected_result, rtol=0.002)


@pytest.mark.parametrize(
    "cable_azimuth, wind_azimuth",
    [
        (0.0, 0.0),
        (0.0, 90.0),
        (0.0, 180.0),
        (90.0, 0.0),
        (90.0, 180.0),
        (30, -80),
        (-60, -80),
        (30, 100),
        (-60, 100),
    ],
)
def test_compute_wind_attack_angle_scalar_bounds(cable_azimuth, wind_azimuth):
    attack_angle = compute_wind_attack_angle(cable_azimuth, wind_azimuth)
    assert 0.0 <= attack_angle <= np.pi / 2.0


def test_compute_wind_attack_angle_array_bounds():
    cable_azimuth = np.linspace(-1080.0, 1080.0, 1001)
    wind_azimuth = np.linspace(1440.0, -1440.0, 1001)

    attack_angle = compute_wind_attack_angle(cable_azimuth, wind_azimuth)

    assert np.all(attack_angle >= 0.0)
    assert np.all(attack_angle <= np.pi / 2.0)


def test_compute_wind_attack_angle_symmetry():
    cable_azimuth = np.array([-720.0, -30.0, 0.0, 45.0, 90.0, 271.0, 810.0])
    wind_azimuth = np.array([1080.0, 15.0, 0.0, 315.0, 180.0, -89.0, -450.0])

    attack_angle_ab = compute_wind_attack_angle(cable_azimuth, wind_azimuth)
    attack_angle_ba = compute_wind_attack_angle(wind_azimuth, cable_azimuth)

    assert np.allclose(attack_angle_ab, attack_angle_ba)


def test_compute_wind_attack_angle_periodicity_360_deg():
    cable_azimuth = np.array([-123.0, -30.5, 0.0, 44.0, 90.0, 278.0, 512.0])
    wind_azimuth = np.array([301.0, -185.0, 0.0, 359.9, 180.0, -92.0, 33.0])

    base = compute_wind_attack_angle(cable_azimuth, wind_azimuth)

    shifted_cable = compute_wind_attack_angle(cable_azimuth + 360.0, wind_azimuth)
    shifted_wind = compute_wind_attack_angle(cable_azimuth, wind_azimuth - 720.0)
    shifted_both = compute_wind_attack_angle(
        cable_azimuth + 1080.0, wind_azimuth - 360.0
    )

    assert np.allclose(base, shifted_cable)
    assert np.allclose(base, shifted_wind)
    assert np.allclose(base, shifted_both)


def test_compute_wind_attack_angle_opposite_angles():
    cable_azimuth = np.array([-123.0, -30.5, 0.0, 44.0, 90.0, 278.0, 512.0])
    wind_azimuth = np.array([301.0, -185.0, 0.0, 359.9, 180.0, -92.0, 33.0])

    base = compute_wind_attack_angle(cable_azimuth, wind_azimuth)

    shifted_cable = compute_wind_attack_angle(cable_azimuth + 180.0, wind_azimuth)
    shifted_wind = compute_wind_attack_angle(cable_azimuth, wind_azimuth - 180.0)
    shifted_both = compute_wind_attack_angle(
        cable_azimuth + 180.0, wind_azimuth - 180.0
    )

    assert np.allclose(base, shifted_cable)
    assert np.allclose(base, shifted_wind)
    assert np.allclose(base, shifted_both)


def test_value_provided_wind_attack_angle():
    dic = set_default_values_scalar()
    dic["wind_attack_angle"] = np.deg2rad(30.0)
    # wind azimuth must be ignored
    dic["wind_azimuth"] = 100.0
    convective_cooling = ConvectiveCooling(**dic)

    assert np.isclose(convective_cooling.wind_attack_angle, np.deg2rad(30.0))

    dic_without_wind_attack_angle = set_default_values_scalar()
    dic_without_wind_attack_angle["wind_azimuth"] = 90.0
    dic_without_wind_attack_angle["cable_azimuth"] = 60.0
    expected_convective_cooling = ConvectiveCooling(
        **dic_without_wind_attack_angle,
    )

    conductor_temperature = 100.0

    assert np.isclose(
        convective_cooling.value(conductor_temperature),
        expected_convective_cooling.value(conductor_temperature),
    )


def test_value_missing_wind_attack_angle_and_wind_azimuth():
    dic = set_default_values_scalar()
    # dic doesn't contain wind_attack_angle
    del dic["wind_azimuth"]

    with pytest.raises(
        ValueError, match="Must provide either wind_attack_angle or wind_azimuth."
    ):
        ConvectiveCooling(**dic)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from math import nan
from unittest.mock import MagicMock

import numpy as np

from thermohl.solver import Solver3T
from thermohl.solver.slv3t import (
    _profile_mom,
    _phi,
    _profile_bim_avg_coeffs,
)


def test_profile_mom_basic():
    ts = np.array([300])
    tc = np.array([400])
    r = np.array([0.5])
    re = np.array([1.0])
    expected = np.array([375])

    result = _profile_mom(ts, tc, r, re)

    np.testing.assert_array_almost_equal(result, expected)


def test_profile_mom_multiple_values():
    ts = np.array([300, 350])
    tc = np.array([400, 450])
    r = np.array([0.5, 0.2])
    re = np.array([1.0, 1.0])
    expected = np.array([375, 446])

    result = _profile_mom(ts, tc, r, re)

    np.testing.assert_array_almost_equal(result, expected)


def test_profile_mom_edge_case_zero_radius():
    ts = np.array([300])
    tc = np.array([400])
    r = np.array([0.0])
    re = np.array([1.0])
    expected = np.array([400])

    result = _profile_mom(ts, tc, r, re)

    np.testing.assert_array_almost_equal(result, expected)


def test_profile_mom_edge_case_equal_temperatures():
    ts = np.array([300])
    tc = np.array([300])
    r = np.array([0.5])
    re = np.array([1.0])
    expected = np.array([300])

    result = _profile_mom(ts, tc, r, re)

    np.testing.assert_array_almost_equal(result, expected)


def test_profile_mom_large_values():
    ts = np.array([1e6])
    tc = np.array([2e6])
    r = np.array([0.5])
    re = np.array([1.0])
    expected = np.array([1.75e6])

    result = _profile_mom(ts, tc, r, re)

    np.testing.assert_array_almost_equal(result, expected)


def test_phi_basic():
    r = 0.6
    ri = 0.5
    re = 1.0
    expected = 0.012559

    result = _phi(r, ri, re)

    np.testing.assert_array_almost_equal(result, expected)


def test_phi_multiple_values():
    r = np.array([0.5, 0.8])
    ri = np.array([0.2, 0.3])
    re = np.array([1.0, 1.0])
    expected = np.array([0.071196, 0.205193])

    result = _phi(r, ri, re)

    np.testing.assert_array_almost_equal(result, expected)


def test_phi_edge_case_zero_radius():
    r = np.array([0.2])
    ri = np.array([0.2])
    re = np.array([1.0])
    expected = np.array([0.0])

    result = _phi(r, ri, re)

    np.testing.assert_array_almost_equal(result, expected)


def test_phi_edge_case_equal_radii():
    r = np.array([0.5])
    ri = np.array([0.5])
    re = np.array([1.0])
    expected = np.array([0.0])

    result = _phi(r, ri, re)

    np.testing.assert_array_almost_equal(result, expected)


def test_phi_large_values():
    r = np.array([1e6])
    ri = np.array([1e5])
    re = np.array([1e7])
    expected = np.array([0.00472])

    result = _phi(r, ri, re)

    np.testing.assert_array_almost_equal(result, expected)


# def test_profile_bim_basic():
#     ts = np.array([300])
#     tc = np.array([400])
#     r = np.array([0.5])
#     ri = np.array([0.2])
#     re = np.array([1.0])
#     expected = np.array([383.55518])
#
#     result = _profile_bim(ts, tc, r, ri, re)
#
#     np.testing.assert_array_almost_equal(result, expected)
#
#
# def test_profile_bim_multiple_values():
#     ts = np.array([300, 350])
#     tc = np.array([400, 450])
#     r = np.array([0.5, 0.2])
#     ri = np.array([0.2, 0.3])
#     re = np.array([1.0, 1.0])
#     expected = np.array([325, 350])
#
#     result = _profile_bim(ts, tc, r, ri, re)
#
#     np.testing.assert_array_almost_equal(result, expected)
#
#
# def test_profile_bim_edge_case_zero_radius():
#     ts = np.array([300])
#     tc = np.array([400])
#     r = np.array([0.0])
#     ri = np.array([0.2])
#     re = np.array([1.0])
#     expected = np.array([400])
#
#     result = _profile_bim(ts, tc, r, ri, re)
#
#     np.testing.assert_array_almost_equal(result, expected)
#
#
# def test_profile_bim_edge_case_equal_temperatures():
#     ts = np.array([300])
#     tc = np.array([300])
#     r = np.array([0.5])
#     ri = np.array([0.2])
#     re = np.array([1.0])
#     expected = np.array([300])
#
#     result = _profile_bim(ts, tc, r, ri, re)
#
#     np.testing.assert_array_almost_equal(result, expected)
#
#
# def test_profile_bim_large_values():
#     ts = np.array([1e6])
#     tc = np.array([2e6])
#     r = np.array([0.5])
#     ri = np.array([0.2])
#     re = np.array([1.0])
#     expected = np.array([1835551.796551])
#
#     result = _profile_bim(ts, tc, r, ri, re)
#
#     np.testing.assert_array_almost_equal(result, expected)


def test_profile_bim_avg_coeffs_basic():
    ri = np.array([0.2])
    re = np.array([1.0])
    expected_a = np.array([0.370445])
    expected_b = np.array([0.831245])

    result_a, result_b = _profile_bim_avg_coeffs(ri, re)

    np.testing.assert_array_almost_equal(result_a, expected_a)
    np.testing.assert_array_almost_equal(result_b, expected_b)


def test_profile_bim_avg_coeffs_multiple_values():
    ri = np.array([0.2, 0.3])
    re = np.array([1.0, 1.0])
    expected_a = np.array([0.370445, 0.279235])
    expected_b = np.array([0.831245, 0.693285])

    result_a, result_b = _profile_bim_avg_coeffs(ri, re)

    np.testing.assert_array_almost_equal(result_a, expected_a)
    np.testing.assert_array_almost_equal(result_b, expected_b)


# Tests Solver3T
def test_morgan_coefficients_basic():
    solver = Solver3T()
    solver.args.outer_diameter = np.array([1.0])
    solver.args.core_diameter = np.array([0.5])
    solver.update()

    expected_c = np.array([0.268951])
    expected_D = np.array([1.0])
    expected_d = np.array([0.5])
    expected_i = np.array([0])

    result_c, result_D, result_d, result_i = solver._morgan_coefficients()

    np.testing.assert_array_almost_equal(result_c, expected_c)
    np.testing.assert_array_almost_equal(result_D, expected_D)
    np.testing.assert_array_almost_equal(result_d, expected_d)
    np.testing.assert_array_equal(result_i, expected_i)


def test_morgan_coefficients_multiple_values():
    solver = Solver3T()
    solver.args.outer_diameter = np.array([1.0, 2.5])
    solver.args.core_diameter = np.array([0.5, 1.0])
    solver.update()

    expected_c = np.array([0.268951, 0.325468])
    expected_D = np.array([1.0, 2.5])
    expected_d = np.array([0.5, 1.0])
    expected_i = np.array([0, 1])

    result_c, result_D, result_d, result_i = solver._morgan_coefficients()

    np.testing.assert_array_almost_equal(result_c, expected_c)
    np.testing.assert_array_almost_equal(result_D, expected_D)
    np.testing.assert_array_almost_equal(result_d, expected_d)
    np.testing.assert_array_equal(result_i, expected_i)


def test_morgan_coefficients_edge_case_zero_d():
    solver = Solver3T()
    solver.args.outer_diameter = np.array([1.0])
    solver.args.core_diameter = np.array([0.0])
    solver.update()

    expected_c = np.array([0.5])
    expected_D = np.array([1.0])
    expected_d = np.array([0.0])
    expected_i = np.array([])

    result_c, result_D, result_d, result_i = solver._morgan_coefficients()

    np.testing.assert_array_almost_equal(result_c, expected_c)
    np.testing.assert_array_almost_equal(result_D, expected_D)
    np.testing.assert_array_almost_equal(result_d, expected_d)
    np.testing.assert_array_equal(result_i, expected_i)


def test_morgan_coefficients_edge_case_equal_D_and_d():
    solver = Solver3T()
    solver.args.outer_diameter = np.array([1.0])
    solver.args.core_diameter = np.array([1.0])
    solver.update()

    expected_c = np.array([nan])
    expected_D = np.array([1.0])
    expected_d = np.array([1.0])
    expected_i = np.array([0])

    result_c, result_D, result_d, result_i = solver._morgan_coefficients()

    np.testing.assert_array_almost_equal(result_c, expected_c)
    np.testing.assert_array_almost_equal(result_D, expected_D)
    np.testing.assert_array_almost_equal(result_d, expected_d)
    np.testing.assert_array_equal(result_i, expected_i)


def test_morgan_coefficients_large_values():
    solver = Solver3T()
    solver.args.outer_diameter = np.array([1e6])
    solver.args.core_diameter = np.array([1e5])
    solver.update()

    expected_c = np.array([0.476742])
    expected_D = np.array([1e6])
    expected_d = np.array([1e5])
    expected_i = np.array([0])

    result_c, result_D, result_d, result_i = solver._morgan_coefficients()

    np.testing.assert_array_almost_equal(result_c, expected_c)
    np.testing.assert_array_almost_equal(result_D, expected_D)
    np.testing.assert_array_almost_equal(result_d, expected_d)
    np.testing.assert_array_equal(result_i, expected_i)


def test_joule_basic():
    solver = Solver3T()
    solver.morgan_coefficients = (
        np.array([0.5]),
        np.array([1.0]),
        np.array([0.5]),
        np.array([0]),
    )
    ts = np.array([300.0])
    tc = np.array([400.0])
    expected = np.array([0.0])

    result = solver.joule(ts, tc)

    np.testing.assert_array_almost_equal(result, expected)


def test_balance_basic():
    solver = Solver3T()
    solver.joule = MagicMock(return_value=np.array([100.0]))
    solver.solar_heating = MagicMock()
    solver.solar_heating.value = MagicMock(return_value=np.array([50.0]))
    solver.convective_cooling = MagicMock()
    solver.convective_cooling.value = MagicMock(return_value=np.array([30.0]))
    solver.radiative_cooling = MagicMock()
    solver.radiative_cooling.value = MagicMock(return_value=np.array([10.0]))
    solver.precipitation_cooling = MagicMock()
    solver.precipitation_cooling.value = MagicMock(return_value=np.array([5.0]))

    ts = np.array([300])
    tc = np.array([400])
    expected = np.array([105.0])

    result = solver.balance(ts, tc)

    np.testing.assert_array_almost_equal(result, expected)
    solver.joule.assert_called_once_with(ts, tc)
    solver.solar_heating.value.assert_called_once_with(ts)
    solver.convective_cooling.value.assert_called_once_with(ts)
    solver.radiative_cooling.value.assert_called_once_with(ts)
    solver.precipitation_cooling.value.assert_called_once_with(ts)


def test_morgan_basic():
    solver = Solver3T()
    solver.morgan_coefficients = (
        np.array([0.5]),
        np.array([1.0]),
        np.array([0.5]),
        np.array([0]),
    )
    solver.joule_heating = MagicMock()
    solver.joule_heating.value = MagicMock(return_value=np.array([100.0]))
    solver.joule = MagicMock(return_value=np.array([100.0]))
    solver.args = MagicMock()
    solver.args.radial_thermal_conductivity = 1.0

    ts = np.array([300])
    tc = np.array([400])
    expected = np.array([92.042])

    result = solver.morgan(ts, tc)

    np.testing.assert_array_almost_equal(result, expected, decimal=3)
    solver.joule.assert_called_once_with(ts, tc)



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Any

import numpy as np

from thermohl import floatArrayLike
from thermohl.power import PowerTerm


class JouleHeating(PowerTerm):
    """Joule heating term."""

    def __init__(
        self,
        transit: floatArrayLike,
        magnetic_coeff: floatArrayLike,
        temperature_coeff_linear: floatArrayLike,
        linear_resistance_dc_20c: floatArrayLike,
        reference_temperature: floatArrayLike = 20.0,
        **kwargs: Any,
    ):
        r"""Init with args.

        If more than one input are numpy arrays, they should have the same size.

        Args:
            transit (float | numpy.ndarray): Transit intensity (A).
            magnetic_coeff (float | numpy.ndarray): Coefficient for magnetic effects (—).
            temperature_coeff_linear (float | numpy.ndarray): Linear resistance augmentation with temperature (K⁻¹).
            linear_resistance_dc_20c (float | numpy.ndarray): Electric resistance per unit length (DC) at 20°C (Ω·m⁻¹).
            reference_temperature (float | numpy.ndarray, optional): Reference temperature (°C). The default is 20.

        """
        self.transit = transit
        self.magnetic_coeff = magnetic_coeff
        self.temp_coeff_linear = temperature_coeff_linear
        self.linear_resistance_dc_20c = linear_resistance_dc_20c
        self.reference_temperature = reference_temperature

    def value(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        r"""Compute joule heating.

        Args:
            conductor_temperature (float | numpy.ndarray): Conductor temperature (°C).

        Returns:
            float | numpy.ndarray: Power term value (W·m⁻¹).

        """
        return (
            self.magnetic_coeff
            * self.linear_resistance_dc_20c
            * (
                1.0
                + self.temp_coeff_linear
                * (conductor_temperature - self.reference_temperature)
            )
            * self.transit**2
        )

    def derivative(self, conductor_temperature: floatArrayLike) -> floatArrayLike:
        r"""Compute joule heating derivative.

        If more than one input are numpy arrays, they should have the same size.

        Args:
            conductor_temperature (float | numpy.ndarray): Conductor temperature (°C).

        Returns:
            float | numpy.ndarray: Power term derivative (W·m⁻¹·K⁻¹).

        """
        return (
            self.magnetic_coeff
            * self.linear_resistance_dc_20c
            * self.temp_coeff_linear
            * self.transit**2
            * np.ones_like(conductor_temperature)
        )



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np
import pytest

from thermohl.power.cigre import ConvectiveCooling
from thermohl.power.cigre import Air


conv_cool_instances = [
    (
        ConvectiveCooling(
            altitude=np.array([100.0]),
            cable_azimuth=np.array([45.0]),
            ambient_temperature=np.array([25.0]),
            wind_speed=np.array([5.0]),
            wind_azimuth=np.array([30.0]),
            outer_diameter=np.array([0.01]),
            roughness_ratio=np.array([0.02]),
        ),
        np.ndarray,
    ),
    (
        ConvectiveCooling(
            altitude=100.0,
            cable_azimuth=45.0,
            ambient_temperature=25.0,
            wind_speed=5.0,
            wind_azimuth=30.0,
            outer_diameter=0.01,
            roughness_ratio=0.02,
        ),
        float,
    ),
]


@pytest.mark.parametrize(
    "convective_cooling, expected_type",
    conv_cool_instances,
    ids=["ConvectiveCooling with arrays", "ConvectiveCooling with scalars"],
)
def test_nu_forced_float_value(convective_cooling, expected_type):
    Tf = 30.0
    nu = 1.5e-5
    expected_result = 17.3425

    result = convective_cooling._nu_forced(Tf, nu)

    np.testing.assert_allclose(result, expected_result, rtol=1e-5)


@pytest.mark.parametrize(
    "convective_cooling, expected_type",
    conv_cool_instances,
    ids=["ConvectiveCooling with arrays", "ConvectiveCooling with scalars"],
)
def test_nu_forced_array_single_value(convective_cooling, expected_type):
    Tf = np.array([30.0])
    nu = np.array([1.5e-5])
    expected_result = np.array([17.3425])

    result = convective_cooling._nu_forced(Tf, nu)

    np.testing.assert_allclose(result, expected_result, rtol=1e-5)


@pytest.mark.parametrize(
    "convective_cooling, expected_type",
    conv_cool_instances,
    ids=["ConvectiveCooling with arrays", "ConvectiveCooling with scalars"],
)
def test_nu_forced_array_values(convective_cooling, expected_type):
    Tf = np.array([30.0, 35.0])
    nu = np.array([1.5e-5, 1.6e-5])
    expected_result = np.array([17.3425, 16.6482])

    result = convective_cooling._nu_forced(Tf, nu)

    np.testing.assert_allclose(result, expected_result, rtol=1e-5)


@pytest.mark.parametrize(
    "convective_cooling, expected_type",
    conv_cool_instances,
    ids=["ConvectiveCooling with arrays", "ConvectiveCooling with scalars"],
)
def test_nu_forced_boundary_conditions(convective_cooling, expected_type):
    Tf = np.array([30.0])
    nu = np.array([1.5e-5])
    convective_cooling.roughness_ratio = np.array([0.05])
    convective_cooling.wind_speed = np.array([100.0])
    expected_result = np.array([115.5214])

    result = convective_cooling._nu_forced(Tf, nu)

    np.testing.assert_allclose(result, expected_result, rtol=1e-5)


@pytest.mark.parametrize(
    "convective_cooling, expected_type",
    conv_cool_instances,
    ids=["ConvectiveCooling with arrays", "ConvectiveCooling with scalars"],
)
def test_nu_natural_single_value(convective_cooling, expected_type):
    air_temperature = 50.0
    Td = 25.0
    nu = Air.kinematic_viscosity(air_temperature)

    result = convective_cooling._nu_natural(air_temperature, Td, nu)

    assert isinstance(result, expected_type)
    np.testing.assert_allclose(result, [3.42404], rtol=1e-4)


@pytest.mark.parametrize(
    "convective_cooling, expected_type",
    conv_cool_instances,
    ids=["ConvectiveCooling with arrays", "ConvectiveCooling with scalars"],
)
def test_nu_natural_array_values(convective_cooling, expected_type):
    air_temperature = np.array([50.0, 60.0, 70.0])
    Td = np.array([25.0, 35.0, 45.0])
    nu = Air.kinematic_viscosity(air_temperature)

    result = convective_cooling._nu_natural(air_temperature, Td, nu)

    assert isinstance(result, np.ndarray)
    assert result.shape == air_temperature.shape
    np.testing.assert_allclose(result, [3.42404, 3.55476, 3.63592], rtol=1e-4)


@pytest.mark.parametrize(
    "convective_cooling, expected_type",
    conv_cool_instances,
    ids=["ConvectiveCooling with arrays", "ConvectiveCooling with scalars"],
)
def test_nu_natural_edge_case(convective_cooling, expected_type):
    air_temperature = 0.0
    Td = 0.0
    nu = Air.kinematic_viscosity(air_temperature)

    result = convective_cooling._nu_natural(air_temperature, Td, nu)

    assert isinstance(result, expected_type)
    assert result == 0.0


@pytest.mark.parametrize(
    "convective_cooling, expected_type",
    conv_cool_instances,
    ids=["ConvectiveCooling with arrays", "ConvectiveCooling with scalars"],
)
def test_nu_natural_high_values(convective_cooling, expected_type):
    air_temperature = 1000.0
    Td = 500.0
    nu = Air.kinematic_viscosity(air_temperature)

    result = convective_cooling._nu_natural(air_temperature, Td, nu)

    assert isinstance(result, expected_type)
    np.testing.assert_allclose(result, [2.18853], rtol=1e-4)


@pytest.mark.parametrize(
    "convective_cooling, expected_type",
    conv_cool_instances,
    ids=["ConvectiveCooling with arrays", "ConvectiveCooling with scalars"],
)
def test_value_single_value(convective_cooling, expected_type):
    conductor_temperature = 75.0

    result = convective_cooling.value(conductor_temperature)

    assert isinstance(result, np.ndarray)
    assert result > 0


@pytest.mark.parametrize(
    "convective_cooling, expected_type",
    conv_cool_instances,
    ids=["ConvectiveCooling with arrays", "ConvectiveCooling with scalars"],
)
def test_value_array_values(convective_cooling, expected_type):
    conductor_temperature = np.array([75.0, 85.0, 95.0])

    result = convective_cooling.value(conductor_temperature)

    assert isinstance(result, np.ndarray)
    assert result.shape == conductor_temperature.shape
    assert np.all(result > 0)


@pytest.mark.parametrize(
    "convective_cooling, expected_type",
    conv_cool_instances,
    ids=["ConvectiveCooling with arrays", "ConvectiveCooling with scalars"],
)
def test_value_edge_case(convective_cooling, expected_type):
    conductor_temperature = 25.0  # Same as ambient temperature

    result = convective_cooling.value(conductor_temperature)

    assert isinstance(result, np.ndarray)
    assert result == 0.0


@pytest.mark.parametrize(
    "convective_cooling, expected_type",
    conv_cool_instances,
    ids=["ConvectiveCooling with arrays", "ConvectiveCooling with scalars"],
)
def test_value_high_values(convective_cooling, expected_type):
    conductor_temperature = 1000.0

    result = convective_cooling.value(conductor_temperature)

    assert isinstance(result, np.ndarray)
    assert result > 0



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timezone
import numpy as np
from numpy import array
from thermohl.solver.entities import TemperatureType, HeatEquationType

from thermohl.solver import rte


def test_solver3t_legacy():
    data = {
        "cpo_nbr_cable": array([1]),
        "cable_azimuth": array([90.0]),
        "wind_speed": array([0]),
        "wind_azimuth": array([45.0]),
        "altitude": array([100]),
        "ambient_temperature": array([20]),
        "transit": array([1500.0]),
        "I0": array([500]),
        "error_code": array([0]),
        "t_core": array([36.164476]),
        "t_surf": array([34.955032]),
        "t_avg": array([35.559754]),
        "temp_error": array([0]),
        "buffer_start": array([1.6150752e09]),
        "outer_diameter": array([0.03105]),
        "core_diameter": array([0.0]),
        "outer_area": array([0.00057]),
        "core_area": array([0.0]),
        "linear_resistance_dc_20c": array([5.83e-05]),
        "m": array([1.539]),
        "heat_capacity": array([900.0]),
        "roughness_ratio": array([0]),
        "radial_thermal_conductivity": array([1]),  # FIXME?
        "temperature_coeff_linear": array([0.0036]),
        "temperature_coeff_quadratic": array([8.0e-07]),
        "magnetic_coeff": array([1.0]),
        "magnetic_coeff_per_a": array([0.0]),
        "solar_absorptivity": array([0.9]),
        "emissivity": array([0.8]),
        "datetime_utc": datetime(2000, 3, 7, 0, tzinfo=timezone.utc),
    }
    Ts0 = [34.955032]
    Tc0 = [36.164476]

    solver = rte(data, heat_equation=HeatEquationType.THREE_TEMPERATURES_LEGACY)
    result = solver.transient_temperature_legacy(
        offset=np.linspace(0, 60, 61),
        surface_temperature_0=Ts0,
        core_temperature_0=Tc0,
        return_power=True,
    )

    print(result)
    assert abs(result[TemperatureType.CORE][-1] - 42) <= 0.5
    assert abs(result[TemperatureType.SURFACE][-1] - 39.9) <= 0.5
    assert abs(result[TemperatureType.AVERAGE][-1] - 40.9) <= 0.5



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

# mypy: ignore-errors
"""Tools to perform Monte Carlo simulations using the thermOHL steady solvers with uncertain input parameters."""

import warnings
from typing import Union, Tuple

import numpy as np
import pandas as pd

from thermohl.solver.entities import (
    TargetType,
    CableLocationListLike,
    CableTypeListLike,
)

from thermohl.utils import depends_on_optional

try:
    import scipy
    from scipy.stats._distn_infrastructure import rv_continuous_frozen as frozen_dist
except ImportError:
    warnings.warn(
        "scipy is not installed. Some functions will not be available.",
        RuntimeWarning,
    )
    frozen_dist = object

from thermohl import distributions
from thermohl import solver
from thermohl import utils


def default_uncertainties() -> dict:
    """
    Get default parameters for uncertainties.

    Returns
    -------
    dict
        Dictionnary of default distributions and parameters.

    """
    return utils.add_default_uncertainties({}, warning=False)


def cummean(values: np.ndarray) -> np.ndarray:
    """Cumulative mean."""
    return np.cumsum(values) / (1 + np.array(range(len(values))))


def cumstd(values: np.ndarray) -> np.ndarray:
    """Cumulative std."""
    return np.sqrt(cummean(values**2) - cummean(values) ** 2)


@depends_on_optional("scipy")
def _get_dist(uncertainty_spec: dict, mean_value: float) -> frozen_dist:
    """Get distribution
    -- based on parameters in dict uncertainty_spec and with mean mean_value"""
    mean = mean_value
    # set std
    standard_deviation = uncertainty_spec["std"]
    if uncertainty_spec["relative_std"]:
        standard_deviation *= mean
    # if std is 0., return uniform
    if standard_deviation == 0.0:
        return scipy.stats.uniform(mean, mean)

    # select ditribution
    if uncertainty_spec["dist"] == "truncnorm":
        lower_bound, upper_bound = uncertainty_spec["min"], uncertainty_spec["max"]
        dist = distributions.truncnorm(
            lower_bound, upper_bound, mean, standard_deviation
        )
    elif uncertainty_spec["dist"] == "vonmises":
        dist = distributions.vonmises(np.deg2rad(mean), np.deg2rad(standard_deviation))
    elif uncertainty_spec["dist"] == "wrapnorm":
        dist = distributions.wrapnorm(np.deg2rad(mean), np.deg2rad(standard_deviation))
    else:
        raise ValueError("Dist keyword not supported")

    return dist


@depends_on_optional("scipy")
def _generate_samples(
    input_params: dict,
    index: int,
    uncertainty_spec: dict,
    num_samples: int,
    include_check: bool = False,
) -> Union[dict, Tuple[dict, pd.DataFrame]]:
    """
    Generate random samples for all input parameters affected with a probability distribution.
    """

    # sample dict
    samples_dict = {}
    # check dataframe
    if include_check:
        columns = [
            "key",
            "dist",
            "mean",
            "std",
            "min",
            "max",
            "s_mean",
            "s_std",
            "s_min",
            "s_max",
            "circular",
        ]
        check_table = pd.DataFrame(
            columns=columns, data=np.zeros((len(input_params), len(columns))) * np.nan
        )
        check_table.loc[:, "circular"] = False

    # loop on dict
    for row_index, key in enumerate(input_params):
        if key not in uncertainty_spec.keys():
            continue
        dist = uncertainty_spec[key]["dist"]
        mean_value = input_params[key][index]
        if dist is None or np.isnan(mean_value):
            sample = mean_value * np.ones((num_samples,), dtype=type(mean_value))
        else:
            sample = _get_dist(uncertainty_spec[key], mean_value).rvs(num_samples)
            if dist == "vonmises" or dist == "wrapnorm":
                sample = np.rad2deg(sample) % 360.0
        samples_dict[key] = sample

        if include_check:
            check_table.loc[row_index, "key"] = key
            check_table.loc[row_index, "dist"] = dist
            check_table.loc[row_index, "mean"] = mean_value
            if "std" in uncertainty_spec[key].keys():
                check_table.loc[row_index, "std"] = uncertainty_spec[key]["std"]
                if uncertainty_spec[key]["relative_std"]:
                    check_table.loc[row_index, "std"] *= mean_value
            if "min" in uncertainty_spec[key].keys():
                check_table.loc[row_index, "min"] = uncertainty_spec[key]["min"]
            if "max" in uncertainty_spec[key].keys():
                check_table.loc[row_index, "max"] = uncertainty_spec[key]["max"]
            if dist in ["vonmises", "wrapnorm"]:
                check_table.loc[row_index, "s_mean"] = scipy.stats.circmean(
                    sample, high=360.0, low=0.0
                )
                check_table.loc[row_index, "s_std"] = scipy.stats.circstd(
                    sample, high=360.0, low=0.0
                )
                check_table.loc[row_index, "circular"] = True
            else:
                check_table.loc[row_index, "s_mean"] = sample.mean()
                check_table.loc[row_index, "s_std"] = sample.std()
            check_table.loc[row_index, "s_min"] = sample.min()
            check_table.loc[row_index, "s_max"] = sample.max()

    if include_check:
        return samples_dict, check_table

    return samples_dict


def _rdict(
    mode: str,
    target: TargetType,
    include_surface: bool,
    include_core: bool,
    include_average: bool,
) -> dict:
    """Code factorization"""
    if mode == "temperature":
        rdc = dict(
            return_core=include_core,
            return_avg=include_average,
            return_power=False,
        )
    elif mode == "intensity":
        rdc = dict(
            target=target,
            return_core=include_core,
            return_avg=include_average,
            return_surf=include_surface,
            return_power=False,
        )
    else:
        raise ValueError("")
    return rdc


def _compute(
    mode: str,
    solver_instance: solver.Solver,
    target_temp: Union[float, np.ndarray],
    return_config: dict,
):
    """Code factorization"""
    if mode == "temperature":
        result = solver_instance.steady_temperature(**return_config)
    elif mode == "intensity":
        result = solver_instance.steady_intensity(target_temp, **return_config)
    else:
        raise ValueError()
    return result


def _steady_uncertainties(
    solver_instance: solver.Solver,
    target_max_temp: Union[float, np.ndarray],
    target_label: TargetType,
    uncertainties: dict,
    num_samples: int,
    include_surface: bool,
    include_core: bool,
    include_average: bool,
    return_raw: bool,
    mode: str = "temperature",
) -> Union[pd.DataFrame, Tuple[pd.DataFrame, list]]:
    """Code factorization"""
    # return dict
    return_config = _rdict(
        mode, target_label, include_surface, include_core, include_average
    )

    # save solver dict
    saved_args = solver_instance.dc

    #  all to max_len size
    num_entries = utils.dict_max_len(solver_instance.dc)
    input_params = utils.extend_to_max_len(solver_instance.dc, num_entries)
    max_temp_vector = target_max_temp * np.ones(
        num_entries,
    )

    # add missing uncertainties parameters
    uncertainty_spec = utils.add_default_uncertainties(uncertainties)

    # init outputs
    raw_results = []
    solver_instance.dc = solver.default_values()
    sample_result = _compute(mode, solver_instance, 99.0, return_config)
    columns = []
    for column_name in sample_result.columns:
        columns.append(column_name + "_mean")
        columns.append(column_name + "_std")
    stats_df = pd.DataFrame(data=np.zeros((num_entries, len(columns))), columns=columns)

    # for each entry, generate sample then compute
    for index in range(num_entries):
        solver_instance.dc = _generate_samples(
            input_params, index, uncertainty_spec, num_samples, include_check=False
        )
        result = _compute(mode, solver_instance, max_temp_vector[index], return_config)
        if return_raw:
            raw_results.append(result)
        mean_values = result.mean()
        std_values = result.std()
        for column_name in result.columns:
            stats_df.loc[index, column_name + "_mean"] = mean_values[column_name]
            stats_df.loc[index, column_name + "_std"] = std_values[column_name]

    # restore solver dict
    solver_instance.dc = saved_args

    if return_raw:
        return stats_df, raw_results
    else:
        return stats_df


def temperature(
    solver_instance: solver.Solver,
    uncertainties: dict = {},
    num_samples: int = 4999,
    return_core: bool = False,
    return_avg: bool = False,
    return_raw: bool = False,
) -> Union[pd.DataFrame, Tuple[pd.DataFrame, list]]:
    """
    Perform Monte Carlo simulation using the steady temperature solver.
    """
    return _steady_uncertainties(
        solver_instance,
        np.nan,
        None,
        uncertainties,
        num_samples,
        None,
        return_core,
        return_avg,
        return_raw,
        mode="temperature",
    )


def intensity(
    solver_instance: solver.Solver,
    target_max_temp: Union[float, np.ndarray],
    target_label: TargetType = TargetType.SURFACE,
    cable_type: CableTypeListLike = None,
    uncertainties: dict = {},
    num_samples: int = 4999,
    return_core: bool = False,
    return_avg: bool = False,
    return_surf: bool = False,
    return_raw: bool = False,
) -> Union[pd.DataFrame, Tuple[pd.DataFrame, list]]:
    """
    Perform Monte Carlo simulation using the steady intensity solver.
    """
    target_label = solver.infer_target_label(target_label, cable_type)
    return _steady_uncertainties(
        solver_instance,
        target_max_temp,
        target_label,
        uncertainties,
        num_samples,
        return_surf,
        return_core,
        return_avg,
        return_raw,
        mode="intensity",
    )


def _diff_method(
    solver_instance: solver.Solver,
    target_max_temp: Union[float, np.ndarray],
    target_label: str,
    uncertainties: dict,
    quantile: float = 0.95,
    return_surf: bool = False,
    return_core: bool = False,
    return_avg: bool = False,
    perturbation_step: float = 1.0e-06,
    mode: str = "temperature",
) -> pd.DataFrame:
    """."""
    # return dict
    return_config = _rdict(
        mode,
        target_label,
        include_surface=return_surf,
        include_core=return_core,
        include_average=return_avg,
    )

    # save solver dict
    saved_args = solver_instance.dc

    #  all to max_len size
    num_entries = utils.dict_max_len(solver_instance.dc)
    input_params = utils.extend_to_max_len(solver_instance.dc, num_entries)

    # add missing uncertainties parameters
    uncertainty_spec = utils.add_default_uncertainties(uncertainties)

    baseline = _compute(mode, solver_instance, target_max_temp, return_config)
    delta_results = baseline * 0.0
    for key in input_params:
        if key not in uncertainty_spec.keys() or uncertainty_spec[key]["dist"] is None:
            continue
        if np.all(np.isnan(input_params[key])):
            continue

        mean_values = input_params[key]
        delta_param = np.zeros_like(mean_values)
        for index in range(num_entries):
            if np.isnan(mean_values[index]):
                delta_param[index] = 0.0
            else:
                dist = _get_dist(uncertainty_spec[key], mean_values[index])
                try:
                    delta_param[index] = (
                        0.5
                        * np.diff(
                            dist.ppf([0.5 * (1 - quantile), 0.5 * (1 + quantile)])
                        )[0]
                    )
                except ValueError:
                    delta_param[index] = 0.0
                if np.isnan(delta_param[index]):
                    delta_param[index] = 0.0
        solver_instance.dc[key] = mean_values + perturbation_step
        perturbed = _compute(mode, solver_instance, target_max_temp, return_config)
        solver_instance.dc[key] = mean_values
        delta_output = np.abs(perturbed - baseline) / perturbation_step
        for column_name in delta_output.columns:
            delta_output.loc[:, column_name] *= delta_param
        delta_results += delta_output
        if np.any(delta_output.isna()):
            print("Nans with key %s" % (key,))

    results = pd.DataFrame()
    for column_name in baseline.columns:
        results.loc[:, column_name] = baseline.loc[:, column_name]
        results.loc[:, column_name + "_delta"] = delta_results.loc[:, column_name]

    # restore solver dict
    solver_instance.dc = saved_args

    return results


def temperature_diff(
    solver_instance: solver.Solver,
    uncertainties: dict,
    quantile: float = 0.95,
    return_core: bool = False,
    return_avg: bool = False,
    perturbation_step: float = 1.0e-06,
) -> pd.DataFrame:
    """."""
    return _diff_method(
        solver_instance,
        np.nan,
        None,
        uncertainties,
        quantile=quantile,
        return_core=return_core,
        return_avg=return_avg,
        perturbation_step=perturbation_step,
        mode="temperature",
    )


def intensity_diff(
    solver_instance: solver.Solver,
    target_max_temp: Union[float, np.ndarray],
    uncertainties: dict,
    target_label: str = None,
    cable_type: CableTypeListLike = None,
    quantile: float = 0.95,
    return_surf: bool = False,
    return_core: bool = False,
    return_avg: bool = False,
    perturbation_step: float = 1.0e-06,
) -> pd.DataFrame:
    """."""
    target_label = solver.infer_target_label(target_label, cable_type)
    return _diff_method(
        solver_instance,
        target_max_temp,
        target_label,
        uncertainties,
        quantile=quantile,
        return_core=return_core,
        return_avg=return_avg,
        return_surf=return_surf,
        perturbation_step=perturbation_step,
        mode="intensity",
    )


def sensitivity(
    solver_instance: solver.Solver,
    target_max_temp: Union[float, np.ndarray],
    uncertainties: dict,
    num_samples: int,
    include_surface: bool,
    include_core: bool,
    include_average: bool,
    target_label: CableLocationListLike = None,
    cable_type: CableTypeListLike = None,
    mode: str = "temperature",
) -> Tuple[list[pd.DataFrame], list[pd.DataFrame]]:
    """
    Perform a sensitivity analysis with Sobol indices (first order and total indices) using the Monte Carlo method.
    """
    target_label = solver.infer_target_label(target_label, cable_type)
    # return dict
    return_config = _rdict(
        mode, target_label, include_surface, include_core, include_average
    )

    # save solver dict
    saved_args = solver_instance.dc

    #  all to max_len size
    num_entries = utils.dict_max_len(solver_instance.dc)
    input_params = utils.extend_to_max_len(solver_instance.dc, num_entries)
    max_temp_vector = target_max_temp * np.ones(
        num_entries,
    )

    # add missing uncertainties parameters
    uncertainty_spec = utils.add_default_uncertainties(uncertainties)

    # init outputs
    first_order_list = []
    total_order_list = []

    # for each entry, compute
    for index in range(num_entries):
        # first sample
        samples_a = _generate_samples(
            input_params, index, uncertainty_spec, num_samples, include_check=False
        )
        solver_instance.dc = samples_a
        result_a = _compute(
            mode, solver_instance, max_temp_vector[index], return_config
        )

        # second sample
        samples_b = _generate_samples(
            input_params, index, uncertainty_spec, num_samples, include_check=False
        )
        solver_instance.dc = samples_b
        result_b = _compute(
            mode, solver_instance, max_temp_vector[index], return_config
        )

        pqs = ((result_a - result_b) ** 2).sum()

        #
        first_order = pd.DataFrame(
            columns=["var"] + result_a.columns.tolist(),
            data=np.zeros((len(uncertainty_spec), 1 + len(result_a.columns))),
        )
        first_order.loc[:, "var"] = uncertainty_spec.keys()
        total_order = pd.DataFrame(
            columns=["var"] + result_a.columns.tolist(),
            data=np.zeros((len(uncertainty_spec), 1 + len(result_a.columns))),
        )
        total_order.loc[:, "var"] = uncertainty_spec.keys()

        # mix samples, run and compute 1st and total indexes
        for row_index, key in enumerate(uncertainty_spec):
            solver_instance.dc = samples_a.copy()
            solver_instance.dc[key] = samples_b[key]
            result_a_mix = _compute(
                mode, solver_instance, max_temp_vector[index], return_config
            )

            solver_instance.dc = samples_b.copy()
            solver_instance.dc[key] = samples_a[key]
            result_b_mix = _compute(
                mode, solver_instance, max_temp_vector[index], return_config
            )

            denom = pqs + ((result_a_mix - result_b_mix) ** 2).sum()
            first_order.iloc[row_index, 1:] = (
                2.0
                * ((result_b_mix - result_b) * (result_a - result_a_mix)).sum()
                / denom
            )
            total_order.iloc[row_index, 1:] = (
                (result_b - result_b_mix) ** 2 + (result_a - result_a_mix) ** 2
            ).sum() / denom

        first_order_list.append(first_order)
        total_order_list.append(total_order)

    # restore solver dict
    solver_instance.dc = saved_args

    return first_order_list, total_order_list



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from typing import Union, List

import numpy as np
import numpy.typing as npt
from datetime import datetime, date

try:
    from ._version import __version__
except ImportError:
    try:
        from importlib.metadata import version

        __version__ = version("thermohl")
    except Exception:
        __version__ = "unknown"

floatArrayLike = Union[float, npt.NDArray[np.float64]]
numberArrayLike = Union[float, int, npt.NDArray[np.float64], npt.NDArray[np.int64]]
strListLike = Union[str, List[str]]
dateListLike = Union[date, List[date]]
datetimeListLike = Union[datetime, List[datetime]]

floatArray = npt.NDArray[np.float64]
intArray = npt.NDArray[np.int64]
numberArray = Union[npt.NDArray[np.float64], npt.NDArray[np.int64]]



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timezone
import numpy as np
import pytest

from thermohl import sun
from thermohl.sun import time_to_float_hours
from thermohl.power import _SRad
from thermohl.power.ieee import SolarHeating


# Tests Classe SRad
@pytest.fixture
def srad():
    clean = [
        -4.22391e01,
        +6.38044e01,
        -1.9220e00,
        +3.46921e-02,
        -3.61118e-04,
        +1.94318e-06,
        -4.07608e-09,
    ]
    indus = [
        +5.31821e01,
        +1.4211e01,
        +6.6138e-01,
        -3.1658e-02,
        +5.4654e-04,
        -4.3446e-06,
        +1.3236e-08,
    ]
    return _SRad(clean, indus)


def test_srad_atmosphere_turbidity_scalar(srad):
    solar_altitude = 30.0
    turbidity = 0.5
    omt = 1.0 - turbidity
    A = omt * srad.clean[6] + turbidity * srad.indus[6]
    B = omt * srad.clean[5] + turbidity * srad.indus[5]
    C = omt * srad.clean[4] + turbidity * srad.indus[4]
    outer_diameter = omt * srad.clean[3] + turbidity * srad.indus[3]
    E = omt * srad.clean[2] + turbidity * srad.indus[2]
    F = omt * srad.clean[1] + turbidity * srad.indus[1]
    G = omt * srad.clean[0] + turbidity * srad.indus[0]
    expected = (
        A * solar_altitude**6
        + B * solar_altitude**5
        + C * solar_altitude**4
        + outer_diameter * solar_altitude**3
        + E * solar_altitude**2
        + F * solar_altitude**1
        + G
    )

    result = srad.atmosphere_turbidity(solar_altitude, turbidity)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_srad_atmosphere_turbidity_array(srad):
    solar_altitude = np.array([30.0, 40.0])
    turbidity = np.array([0.5, 0.7])
    omt = 1.0 - turbidity
    A = omt * srad.clean[6] + turbidity * srad.indus[6]
    B = omt * srad.clean[5] + turbidity * srad.indus[5]
    C = omt * srad.clean[4] + turbidity * srad.indus[4]
    outer_diameter = omt * srad.clean[3] + turbidity * srad.indus[3]
    E = omt * srad.clean[2] + turbidity * srad.indus[2]
    F = omt * srad.clean[1] + turbidity * srad.indus[1]
    G = omt * srad.clean[0] + turbidity * srad.indus[0]
    expected = (
        A * solar_altitude**6
        + B * solar_altitude**5
        + C * solar_altitude**4
        + outer_diameter * solar_altitude**3
        + E * solar_altitude**2
        + F * solar_altitude**1
        + G
    )

    result = srad.atmosphere_turbidity(solar_altitude, turbidity)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


def test_srad_call_scalar(srad):
    latitude = 45.0
    altitude = 1000.0
    cable_azimuth = 180.0
    turbidity = 0.5
    datetime_utc = datetime(2000, 6, 21, 12, tzinfo=timezone.utc)
    date = datetime_utc.date()
    hour = np.array([time_to_float_hours(datetime_utc.time())])
    sa = sun.solar_altitude(latitude, date, hour)
    sz = sun.solar_azimuth(latitude, date, hour)
    th = np.arccos(np.cos(sa) * np.cos(sz - cable_azimuth))
    K = 1.0 + 1.148e-04 * altitude - 1.108e-08 * altitude**2
    Q = srad.atmosphere_turbidity(np.rad2deg(sa), turbidity)
    expected = K * Q * np.sin(th)
    expected = np.where(expected > 0.0, expected, 0.0)

    result = srad(latitude, altitude, cable_azimuth, turbidity, datetime_utc)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


def test_srad_call_array(srad):
    latitude = np.array([45.0, 50.0])
    altitude = np.array([1000.0, 2000.0])
    cable_azimuth = np.array([180.0, 190.0])
    turbidity = np.array([0.5, 0.7])
    datetime_utc = [
        datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
        datetime(2000, 7, 22, 13, tzinfo=timezone.utc),
    ]
    date = [d.date() for d in datetime_utc]
    hour = np.array([time_to_float_hours(d.time()) for d in datetime_utc])
    sa = sun.solar_altitude(latitude, date, hour)
    sz = sun.solar_azimuth(latitude, date, hour)
    th = np.arccos(np.cos(sa) * np.cos(sz - cable_azimuth))
    K = 1.0 + 1.148e-04 * altitude - 1.108e-08 * altitude**2
    Q = srad.atmosphere_turbidity(np.rad2deg(sa), turbidity)
    expected = K * Q * np.sin(th)
    expected = np.where(expected > 0.0, expected, 0.0)

    result = srad(latitude, altitude, cable_azimuth, turbidity, datetime_utc)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


# Tests Class SolarHeating
solar_heating_instances = [
    SolarHeating(
        latitude=np.array([45.0, 50.0]),
        altitude=np.array([1000.0, 2000.0]),
        cable_azimuth=np.array([180.0, 190.0]),
        turbidity=np.array([0.5, 0.7]),
        datetime_utc=[
            datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
            datetime(2000, 7, 22, 13, tzinfo=timezone.utc),
        ],
        outer_diameter=np.array([0.01, 0.02]),
        solar_absorptivity=np.array([0.9, 0.8]),
        measured_solar_irradiance=np.array([800.0, 900.0]),
    ),
    SolarHeating(
        latitude=45.0,
        altitude=1000.0,
        cable_azimuth=180.0,
        turbidity=0.5,
        datetime_utc=datetime(2000, 6, 21, 12, tzinfo=timezone.utc),
        outer_diameter=0.01,
        solar_absorptivity=0.9,
        measured_solar_irradiance=800.0,
    ),
]


@pytest.mark.parametrize(
    "solar_heating",
    solar_heating_instances,
    ids=[
        "SolarHeating with arrays",
        "SolarHeating with scalars",
    ],
)
def test_solar_heating_value_scalar(solar_heating):
    conductor_temperature = 50.0
    expected = (
        solar_heating.solar_absorptivity
        * solar_heating.solar_irradiance
        * solar_heating.outer_diameter
    )

    result = solar_heating.value(conductor_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


@pytest.mark.parametrize(
    "solar_heating",
    solar_heating_instances,
    ids=[
        "SolarHeating with arrays",
        "SolarHeating with scalars",
    ],
)
def test_solar_heating_value_array(solar_heating):
    conductor_temperature = np.array([50.0, 60.0])
    expected = (
        solar_heating.solar_absorptivity
        * solar_heating.solar_irradiance
        * solar_heating.outer_diameter
    )

    result = solar_heating.value(conductor_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"


@pytest.mark.parametrize(
    "solar_heating",
    solar_heating_instances,
    ids=[
        "SolarHeating with arrays",
        "SolarHeating with scalars",
    ],
)
def test_solar_heating_derivative_scalar(solar_heating):
    conductor_temperature = 50.0
    expected = 0.0

    result = solar_heating.derivative(conductor_temperature)

    assert np.isclose(result, expected), f"Expected {expected}, but got {result}"


@pytest.mark.parametrize(
    "solar_heating",
    solar_heating_instances,
    ids=[
        "SolarHeating with arrays",
        "SolarHeating with scalars",
    ],
)
def test_solar_heating_derivative_array(solar_heating):
    conductor_temperature = np.array([50.0, 60.0])
    expected = np.array([0.0, 0.0])

    result = solar_heating.derivative(conductor_temperature)

    assert np.allclose(result, expected), f"Expected {expected}, but got {result}"



# SPDX-FileCopyrightText: 2025 RTE (https://www.rte-france.com)
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np

from thermohl import floatArrayLike

_zerok = 273.15


def kelvin(t: floatArrayLike) -> floatArrayLike:
    return t + _zerok


class Air:
    """`Wikipedia <https://fr.wikipedia.org/wiki/Air> models."""

    @staticmethod
    def volumic_mass(
        air_temperature: floatArrayLike, altitude: floatArrayLike = 0.0
    ) -> floatArrayLike:
        r"""
        Compute air volumic mass.

        If both inputs are numpy arrays, they should have the same size.

        Args:
            air_temperature (float | numpy.ndarray): Air temperature (in Celsius).
            altitude (float | numpy.ndarray, optional): Altitude above sea-level. The default is 0.

        Returns:
            float | numpy.ndarray: Volumic mass in kg·m⁻³.

        """
        air_temperature_k = kelvin(air_temperature)
        return (
            1.292
            * _zerok
            * np.exp(-3.42e-02 * altitude / air_temperature_k)
            / air_temperature_k
        )

    @staticmethod
    def dynamic_viscosity(air_temperature: floatArrayLike) -> floatArrayLike:
        r"""Compute air dynamic viscosity.

        Args:
            air_temperature (float | numpy.ndarray): Air temperature (in Celsius)

        Returns:
            float | numpy.ndarray: Dynamic viscosity in kg·m⁻¹·s⁻¹.

        """
        air_temperature_k = kelvin(air_temperature)
        return (
            8.8848e-15 * air_temperature_k**3
            - 3.2398e-11 * air_temperature_k**2
            + 6.2657e-08 * air_temperature_k
            + 2.3543e-06
        )

    @staticmethod
    def kinematic_viscosity(
        air_temperature: floatArrayLike, altitude: floatArrayLike = 0.0
    ) -> floatArrayLike:
        r"""Compute air kinematic viscosity.

        Args:
            air_temperature (float | numpy.ndarray): Air temperature (in Celsius)
            altitude (float | numpy.ndarray, optional): Altitude above sea-level. The default is 0.

        Returns:
            float | numpy.ndarray: Kinematic viscosity in m²·s⁻¹.

        """
        return Air.dynamic_viscosity(air_temperature) / Air.volumic_mass(
            air_temperature, altitude=altitude
        )

    @staticmethod
    def thermal_conductivity(air_temperature: floatArrayLike) -> floatArrayLike:
        r"""Compute air thermal conductivity.

        The output is valid for input in [-150, 1300] range (in Celsius)

        Args:
            air_temperature (float | numpy.ndarray): Air temperature (in Celsius)

        Returns:
            float | numpy.ndarray: Thermal conductivity in W·m⁻¹·K⁻¹.

        """
        air_temperature_k = kelvin(air_temperature)
        return (
            1.5207e-11 * air_temperature_k**3
            - 4.8570e-08 * air_temperature_k**2
            + 1.0184e-04 * air_temperature_k
            - 3.9333e-04
        )
